import re


class FuzzySearcher(object):

    def __init__(self, char_match_threshold=0.5, ngram_threshold=0.5, levenshtein_threshold=0.5, max_length_variance=1):
        self.char_match_threshold = char_match_threshold
        self.ngram_threshold = ngram_threshold
        self.levenshtein_threshold = levenshtein_threshold
        self.perform_strip_suffix = True
        self.max_length_variance = max_length_variance

    def enable_strip_suffix(self):
        self.perform_strip_suffix = True

    def disable_strip_suffix(self):
        self.perform_strip_suffix = False

    #################################
    # String manipulation functions #
    #################################

    def make_ngrams(self, term, n):
        term = "#{t}#".format(t=term)
        max_start = len(term) - n + 1
        return [term[start:start+n] for start in range(0, max_start)]

    def strip_suffix(self, match):
        if match[-2] in [" ", ","]:
            match = match[:-2]
        elif match[-2:] in [", ", ". ", "? ", ".f"]:
            match = match[:-2]
        elif match[-1] in [" ", ",", "."]:
            match = match[:-1]
        return match

    #####################################
    # Term similarity scoring functions #
    #####################################

    def score_levenshtein_distance(self, s1, s2):
        if len(s1) > len(s2):
            s1, s2 = s2, s1
        distances = range(len(s1) + 1)
        for i2, c2 in enumerate(s2):
            distances_ = [i2+1]
            for i1, c1 in enumerate(s1):
                if c1 == c2:
                    distances_.append(distances[i1])
                else:
                    distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
            distances = distances_
        return distances[-1]

    def score_char_overlap(self, term1, term2):
        num_char_matches = 0
        for char in term2:
            if char in term1:
                term1 = term1.replace(char, "", 1)
                num_char_matches += 1
        return num_char_matches

    def score_ngram_overlap(self, term1, term2, ngram_size):
        term1_ngrams = self.make_ngrams(term1, ngram_size)
        term2_ngrams = self.make_ngrams(term2, ngram_size)
        overlap = 0
        for ngram in term1_ngrams:
            if ngram in term2_ngrams:
                term2_ngrams.pop(term2_ngrams.index(ngram))
                overlap += 1
        return overlap

    def score_char_overlap_ratio(self, term1, term2):
        max_overlap = len(term1)
        overlap = self.score_char_overlap(term1, term2)
        return overlap / max_overlap

    def score_ngram_overlap_ratio(self, term1, term2, ngram_size):
        max_overlap = len(self.make_ngrams(term1, ngram_size))
        overlap = self.score_ngram_overlap(term1, term2, ngram_size)
        return overlap / max_overlap

    def score_levenshtein_distance_ratio(self, term1, term2):
        max_distance = max(len(term1), len(term2))
        distance = self.score_levenshtein_distance(term1, term2)
        return 1 - distance / max_distance

    #################################
    # Candidate filtering functions #
    #################################

    def filter_char_match_candidates(self, candidates, match_term):
        if len(candidates) == 0:
            return candidates
        if isinstance(candidates[0], str):
            return [candidate for candidate in candidates if self.score_char_overlap_ratio(candidate, match_term) >= self.char_match_threshold]
        elif isinstance(candidates[0], dict):
            return [candidate for candidate in candidates if self.score_char_overlap_ratio(candidate["match_string"], match_term) >= self.char_match_threshold]

    def filter_ngram_candidates(self, candidates, match_term, ngram_size):
        if len(candidates) == 0:
            return candidates
        if isinstance(candidates[0], str):
            return [candidate for candidate in candidates if self.score_ngram_overlap_ratio(candidate, match_term, ngram_size) >= self.ngram_threshold]
        elif isinstance(candidates[0], dict):
            return [candidate for candidate in candidates if self.score_ngram_overlap_ratio(candidate["match_string"], match_term, ngram_size) >= self.ngram_threshold]

    def filter_levenshtein_candidates(self, candidates, match_term):
        if len(candidates) == 0:
            return candidates
        if isinstance(candidates[0], str):
            return [candidate for candidate in candidates if self.score_levenshtein_distance_ratio(candidate, match_term) >= self.levenshtein_threshold]
        elif isinstance(candidates[0], dict):
            return [candidate for candidate in candidates if self.score_levenshtein_distance_ratio(candidate["match_string"], match_term) >= self.levenshtein_threshold]

    def filter_candidates(self, candidates, keyword, ngram_size=2):
        if len(candidates) == 0:
            return candidates
        char_match_candidates = self.filter_char_match_candidates(candidates, keyword)
        ngram_candidates = self.filter_ngram_candidates(char_match_candidates, keyword, ngram_size)
        return self.filter_levenshtein_candidates(ngram_candidates, keyword)

    def rank_candidates(self, candidates, keyword, ngram_size=2):
        total_scores = []
        for candidate in candidates:
            match_string = None
            if isinstance(candidate, str):
                match_string = candidate
            elif isinstance(candidate, dict):
                match_string = candidate["match_string"]
            else:
                raise ValueError("candidate must string or fuzzy match object")
            score = {
                "candidate": candidate,
                "char": self.score_char_overlap_ratio(match_string, keyword),
                "ngram": self.score_ngram_overlap_ratio(match_string, keyword, ngram_size),
                "levenshtein": self.score_levenshtein_distance_ratio(match_string, keyword),
            }
            score["total"] = score["char"] + score["ngram"] + score["levenshtein"]
            total_scores += [score]
        return sorted(total_scores, key=lambda x: x["total"], reverse=True)

    ##########################################
    # Functions for finding patterns in text #
    ##########################################

    """
    Find candidate matches that start with the same initial character as the search term and
    that are within max_length_variance different in length.
    """
    def find_term_matches(self, text, term, max_length_variance=None, use_word_boundaries=False):
        if not max_length_variance:
            max_length_variance = self.max_length_variance
        initial = term[0]
        length_range = {"min": len(term[1:]) - max_length_variance, "max": len(term[1:]) + max_length_variance}
        if initial in ["[","]", "*", "(",")", "."]:
            initial = "\\" + initial
        pattern = initial + ".{" + str(length_range["min"]) + "," + str(length_range["max"]) + "}"
        if use_word_boundaries:
            pattern = r"\b" + pattern + r"\b"
        try:
            return [create_term_match(re_match, term) for re_match in re.finditer(pattern, text)]
        except TypeError:
            print("\n\nERROR\n\n")
            print("text:", text)
            print("term:", term)
            print("pattern:", pattern)
            raise

    def find_start_candidates(self, text, term, use_word_boundaries):
        """Find candidate matches that start with the same initial character as the search term."""
        candidates = []
        for match in self.find_term_matches(text, term, use_word_boundaries=use_word_boundaries):
            if self.perform_strip_suffix:
                match["match_string"] = self.strip_suffix(match["match_string"])
            candidates.append(match)
        return candidates

    def find_candidates(self, text, keyword, ngram_size=2, use_word_boundaries=False):
        """Find candidate matches that start with the same initial character as the search term
        and filter them based on default thresholds for character overlap, ngram overlap and
        levenshtein distance."""
        candidates = self.find_start_candidates(text, keyword, use_word_boundaries)
        candidates = self.filter_char_match_candidates(candidates, keyword)
        candidates = self.filter_ngram_candidates(candidates, keyword, ngram_size)
        candidates = self.filter_levenshtein_candidates(candidates, keyword)
        return candidates


def create_term_match(re_match, term):
    return {
        "match_term": term,
        "match_string": re_match.group(0),
        "match_offset": re_match.start()
    }







from typing import List, Generator
from itertools import combinations


#################################
# String manipulation functions #
#################################

def make_ngrams(text: str, n: int) -> List[str]:
    """Turn a term string into a list of ngrams of size n

    :param text: a text string
    :type text: str
    :param n: the ngram size
    :type n: int
    :return: a list of ngrams
    :rtype: List[str]"""
    if not isinstance(text, str):
        raise TypeError('text must be a string')
    if not isinstance(n, int):
        raise TypeError('n must be a positive integer')
    if n < 1:
        raise ValueError('n must be a positive integer')
    if n > len(text):
        return []
    text = "#{t}#".format(t=text)
    max_start = len(text) - n + 1
    return [text[start:start + n] for start in range(0, max_start)]


#####################################
# Term similarity scoring functions #
#####################################


def score_ngram_overlap(term1: str, term2: str, ngram_size: int):
    """Score the number of overlapping ngrams between two terms

    :param term1: a first term string
    :type term1: str
    :param term2: a second term string
    :type term2: str
    :param ngram_size: the character ngram size
    :type ngram_size: int
    :return: the number of overlapping ngrams
    :rtype: int
    """
    term1_ngrams = make_ngrams(term1, ngram_size)
    term2_ngrams = make_ngrams(term2, ngram_size)
    overlap = 0
    for ngram in term1_ngrams:
        if ngram in term2_ngrams:
            term2_ngrams.pop(term2_ngrams.index(ngram))
            overlap += 1
    return overlap


def score_ngram_overlap_ratio(term1, term2, ngram_size):
    """Score the number of overlapping ngrams between two terms as proportion of the length
    of the first term

    :param term1: a term string
    :type term1: str
    :param term2: a term string
    :type term2: str
    :param ngram_size: the character ngram size
    :type ngram_size: int
    :return: the number of overlapping ngrams
    :rtype: int
    """
    max_overlap = len(make_ngrams(term1, ngram_size))
    overlap = score_ngram_overlap(term1, term2, ngram_size)
    return overlap / max_overlap


def score_char_overlap_ratio(term1, term2):
    """Score the number of overlapping characters between two terms as proportion of the length
    of the first term

    :param term1: a term string
    :type term1: str
    :param term2: a term string
    :type term2: str
    :return: the number of overlapping ngrams
    :rtype: int
    """
    max_overlap = len(term1)
    overlap = score_char_overlap(term1, term2)
    return overlap / max_overlap


def score_char_overlap(term1: str, term2: str) -> int:
    """Count the number of overlapping character tokens in two strings.

    :param term1: a term string
    :type term1: str
    :param term2: a term string
    :type term2: str
    :return: the number of overlapping ngrams
    :rtype: int
    """
    num_char_matches = 0
    for char in term2:
        if char in term1:
            term1 = term1.replace(char, "", 1)
            num_char_matches += 1
    return num_char_matches


def score_levenshtein_similarity_ratio(term1, term2):
    """Score the levenshtein similarity between two terms

    :param term1: a term string
    :type term1: str
    :param term2: a term string
    :type term2: str
    :return: the number of overlapping ngrams
    :rtype: int
    """
    max_distance = max(len(term1), len(term2))
    distance = score_levenshtein_distance(term1, term2)
    return 1 - distance / max_distance


def score_levenshtein_distance(term1: str, term2: str) -> int:
    """Calculate Levenshtein distance between two string.

    :param term1: a term string
    :type term1: str
    :param term2: a term string
    :type term2: str
    :return: the number of overlapping ngrams
    :rtype: int
    """
    if len(term1) > len(term2):
        term1, term2 = term2, term1
    distances = range(len(term1) + 1)
    for i2, c2 in enumerate(term2):
        distances_ = [i2 + 1]
        for i1, c1 in enumerate(term1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]


class SkipGram:

    def __init__(self, skipgram_string: str, offset: int, skipgram_length: int):
        self.string = skipgram_string
        self.offset = offset
        self.length = skipgram_length


def insert_skips(window: str, skipgram_combinations: List[List[int]]):
    """For a given skip gram window, return all skip grams for a given configuration."""
    for combination in skipgram_combinations:
        # prev_index = 0
        skip_gram = window[0]
        try:
            for index in combination:
                # if index - prev_index > 1:
                #     skip_gram += "_"
                skip_gram += window[index]
                # prev_index = index
            yield skip_gram, combination[-1] + 1
        except IndexError:
            pass


def text2skipgrams(text: str, ngram_size: int = 2, skip_size: int = 2) -> Generator[SkipGram, None, None]:
    """Turn a text string into a list of skipgrams.

    :param text: an text string
    :type text: str
    :param ngram_size: an integer indicating the number of characters in the ngram
    :type ngram_size: int
    :param skip_size: an integer indicating how many skip characters in the ngrams
    :type skip_size: int
    :return: An iterator returning tuples of skip_gram and offset
    :rtype: Generator[tuple]"""
    if ngram_size <= 0 or skip_size < 0:
        raise ValueError('ngram_size must be a positive integer, skip_size must be a positive integer or zero')
    indexes = [i for i in range(0, ngram_size+skip_size)]
    skipgram_combinations = [combination for combination in combinations(indexes[1:], ngram_size-1)]
    for offset in range(0, len(text)-1):
        window = text[offset:offset+ngram_size+skip_size]
        for skipgram, skipgram_length in insert_skips(window, skipgram_combinations):
            yield SkipGram(skipgram, offset, skipgram_length)


non_word_affixes_2 = {
    ". ", ", ", "! ", "? ",
    " (", ") ", ").", ")!", "),", ")?",
    " [", "] ", "].", "]!", "],", "]?",
}
non_word_affixes_1 = {
    " ", ".", ",", "!", "?",
}


def get_non_word_prefix(string: str) -> str:
    """Check if a string has a non-word prefix and return it.

    :param string: the string from which the prefix is to be return
    :type string: str
    :return: the non-word prefix
    :rtype: str
    """
    if string[:2] in non_word_affixes_2:
        return string[:2]
    elif string[:1] in non_word_affixes_1:
        return string[:1]
    else:
        return ""


def get_non_word_suffix(string: str) -> str:
    """Check if a string has a non-word suffix and return it.

    :param string: the string from which the suffix is to be return
    :type string: str
    :return: the non-word suffix
    :rtype: str
    """
    if string[-2:] in non_word_affixes_2:
        return string[-2:]
    elif string[-1:] in non_word_affixes_1:
        return string[-1:]
    else:
        return ""


def strip_prefix(string: str) -> str:
    """Strip non-word prefix from string ending.

    :param string: the string from which the prefix is to be stripped
    :type string: str
    :return: the stripped string
    :rtype: str
    """
    if len(string) <= 2:
        pass
    elif string[:2] in non_word_affixes_2:
        string = string[2:]
    elif string[1] in [" ", ","]:
        string = string[2:]
    elif string[0] in [" ", ","]:
        string = string[1:]
    return string


def strip_suffix(string: str) -> str:
    """Strip non-word suffix from string ending.

    :param string: the string from which the suffix is to be stripped
    :type string: str
    :return: the stripped string
    :rtype: str
    """
    if len(string) <= 2:
        pass
    elif string[-2] in [" ", ","]:
        string = string[:-2]
    elif string[-2:] in [", ", ". ", "! ", "? "]:
        string = string[:-2]
    elif string[-1] in [" ", ","]:
        string = string[:-1]
    return string



from typing import Dict, List, Union

from fuzzy_search.fuzzy_match import PhraseMatch
from fuzzy_search.fuzzy_phrase_model import PhraseModel
from fuzzy_search.fuzzy_template import FuzzyTemplate, FuzzyTemplateElement, FuzzyTemplateGroupElement
from fuzzy_search.fuzzy_context_searcher import FuzzyContextSearcher


DEBUG = False


def share_label(object1: Union[PhraseMatch, FuzzyTemplateElement],
                object2: Union[PhraseMatch, FuzzyTemplateElement]) -> bool:
    """Check if two fuzzy objects (phrase matches of template elements) share at least one label.

    :param object1: the first object to compare
    :type object1: Union[PhraseMatch, FuzzyTemplateElement]
    :param object2: the second object to compare
    :type object2: Union[PhraseMatch, FuzzyTemplateElement]
    :return: boolean value indicating that the two objects share a label
    :rtype: bool
    """
    label1 = set(object1.label) if isinstance(object1.label, list) else {object1.label}
    label2 = set(object2.label) if isinstance(object2.label, list) else {object2.label}
    # print("set 1:", label1)
    # print("set 2:", label2)
    return len(label1.intersection(label2)) > 0


def get_phrase_match_list_labels(phrase_matches: List[PhraseMatch]) -> List[str]:
    """Return a list of all the labels of a list of phrase matches.

    :param phrase_matches: a list of phrase matches
    :type phrase_matches: List[PhraseMatch]
    :return: a list of phrase match labels
    :rtype: List[str]
    """
    labels: List[str] = []
    for phrase_match in phrase_matches:
        labels += phrase_match.label if isinstance(phrase_match.label, list) else [phrase_match.label]
    return labels


def has_required_matches(phrase_matches: List[PhraseMatch], template: FuzzyTemplate) -> bool:
    """Check if list of phrase matches contain all required labels of a template.

    :param phrase_matches: a list of phrase matches
    :type phrase_matches: List[PhraseMatch]
    :param template: a fuzzy template to use for searching
    :type template: FuzzyTemplate
    :return: a True value only if all required labels have at least one match
    """
    phrase_match_labels = get_phrase_match_list_labels(phrase_matches)
    for required_label in template.get_required_labels():
        if required_label not in phrase_match_labels:
            return False
    return True


class TemplateMatch:

    def __init__(self, template: FuzzyTemplate, phrase_matches: List[PhraseMatch],
                 template_sequence: Dict[str, any]):
        """A match object for a given template, with a list of phrase matches that fill the template elements.

        :param template: the template that is matched
        :type template: FuzzyTemplate
        :param phrase_matches: the phrase matches that correspond to the template elements
        :type phrase_matches: List[PhraseMatch]
        :param template_sequence: a template sequence mapping each phrase match to the corresponding template labels
        :type template_sequence: Dict[str, any]
        """
        self.template = template
        self.phrase_matches = phrase_matches
        self.template_sequence = template_sequence
        self.element_matches = get_sequence_label_element_matches(template_sequence)

    def __repr__(self):
        return f"{self.__class__.__name__}(template={self.template}, element_matches={self.element_matches})"


def find_next_element_start_index(phrase_matches: List[PhraseMatch],
                                  template_element: FuzzyTemplateElement,
                                  template_start_index: int) -> int:
    """Find the next phrase match that matches a template element, from a given starting point
    in a list of phrase matches.

    :param phrase_matches: a list of phrase matches to be tested against a template element
    :type phrase_matches: List[PhraseMatch]
    :param template_element: a template element to test the phrase matches against
    :type template_element: FuzzyTemplateElement
    :param template_start_index: the index in the phrase list to start the matching process
    :type template_start_index: int
    :return: the index in the phrase list where the template element matches
    :rtype: int
    """
    # print("finding start for template element:", template_element.label, template_element.type, template_start_index)
    # if isinstance(template_element, FuzzyTemplateGroupElement):
        # print(template_element.group_element_labels)
    # print("finding start index:")
    for phrase_match in phrase_matches[template_start_index:]:
        if share_label(template_element, phrase_match):
            # print(phrase_matches.index(phrase_match), "phrase_match label:", phrase_match.label)
            # print(phrase_matches.index(phrase_match), "shared label")
            # print("returning start for template element:", template_element.label, template_element.type,
            #       template_start_index)
            return phrase_matches.index(phrase_match)
    return -1


def find_next_element_end_index(phrase_matches: List[PhraseMatch],
                                template_element: FuzzyTemplateElement,
                                element_start_index: int) -> int:
    """Find the next phrase match that doesn't match a template element, from a given starting point
    in a list of phrase matches.

    :param phrase_matches: a list of phrase matches to be tested against a template element
    :type phrase_matches: List[PhraseMatch]
    :param template_element: a template element to test the phrase matches against
    :type template_element: FuzzyTemplateElement
    :param element_start_index: the index in the phrase list where the template elements first matches the template
    :type element_start_index: int
    :return: the index in the phrase list where the template element stops matching
    :rtype: int
    """
    if element_start_index == -1:
        return -1
    # print("template element finding end:", template_element.label, template_element.type)
    for phrase_match in phrase_matches[element_start_index:]:
        if not share_label(template_element, phrase_match):
            # this is the first phrase match after the start index that does not fit
            # the template element, so the sequence ends here
            # print("element end index:", phrase_matches.index(phrase_match))
            return phrase_matches.index(phrase_match)
    # if all phrases matches after element start fit the element's label(s),
    # the end index is the end of the phrase matches
    return len(phrase_matches)


def initialize_sequence(element: FuzzyTemplateElement, start_index: int, end_index: int) -> Dict[str, any]:
    return {
        "element_label": element.label,
        "element_type": element.type,
        "element": element,
        "start": start_index,
        "end": end_index,
        "phrase_matches": [],
        "contains_required": False,
        "element_sequences": []
    }


def get_sequence_label_element_matches(template_sequence: Dict[str, any]) -> List[Dict[str, any]]:
    if template_sequence["element_type"] == "label":
        # print("label:", template_sequence["element_label"], template_sequence["start"], template_sequence["end"])
        return [{"label": template_sequence["element_label"], "phrase_matches": template_sequence["phrase_matches"]}]
    else:
        # print("group:", template_sequence["element_label"], template_sequence["start"], template_sequence["end"])
        template_phrase_matches: List[Dict[str, any]] = []
        for element_sequence in template_sequence["element_sequences"]:
            element_phrase_matches = get_sequence_label_element_matches(element_sequence)
            for element_phrase_match in element_phrase_matches:
                template_phrase_matches.append(element_phrase_match)
                if template_sequence["element_label"]:
                    if "label_groups" not in element_phrase_match:
                        element_phrase_match["label_groups"] = []
                    element_phrase_match["label_groups"].append(template_sequence["element_label"])
                # print("\telement:", element_phrase_match)
        return template_phrase_matches


def find_next_ordered_group_match_sequence(phrase_matches: List[PhraseMatch],
                                           template_group: FuzzyTemplateGroupElement,
                                           template_start_index: int) -> Union[None, Dict[str, any]]:
    """Find the next sequence of phrase matches that match an ordered template group element, from a given
    starting point in the list of phrase matches. This function returns None if the template doesn't match.

    :param phrase_matches: a list of phrase matches to be tested against a template element
    :type phrase_matches: List[PhraseMatch]
    :param template_group: a template group element to test the phrase matches against
    :type template_group: FuzzyTemplateGroupElement
    :param template_start_index: the index in the phrase list to start the matching process
    :type template_start_index: int
    :return: a sequence with start and end indexes in the list of phrase matches that match the template group
    :rtype: Union[None, Dict[str, any]]
    """
    # find first phrase match that belongs to first element of group
    group_sequence = initialize_sequence(template_group, template_start_index, template_start_index)
    element_sequences = []
    if DEBUG:
        print("ordered template group label:", template_group.label, "start index:", template_start_index)
    for group_element in template_group.elements:
        if DEBUG:
            print("ordered group element label:", group_element.label)
        if isinstance(group_element, FuzzyTemplateGroupElement):
            element_sequence = find_next_group_match_sequence(phrase_matches, group_element, group_sequence["end"])
            if DEBUG:
                print("returned sequence within ordered template:", group_element.label, element_sequence)
                if element_sequence:
                    group_sequence["element_sequences"].append(element_sequence)
        else:
            element_start_index = find_next_element_start_index(phrase_matches,
                                                                group_element, group_sequence["start"])
            element_end_index = find_next_element_end_index(phrase_matches,
                                                            group_element, element_start_index)
            if DEBUG:
                print(group_element.label, "element_start_index:", element_start_index,
                      "element_end_index:", element_end_index)
            element_sequence = initialize_sequence(group_element, element_start_index, element_end_index)
            element_sequence["phrase_matches"] = phrase_matches[element_start_index:element_end_index]
        element_sequences.append(element_sequence)
        if element_sequence:
            if DEBUG:
                print(f"ordered group element {group_element.label} indexes:", element_sequence["start"], element_sequence["end"])
                print("\tgroup start,end:", group_sequence["start"], group_sequence["end"])
        if (not element_sequence or element_sequence["start"] == -1) and not group_element.required:
            if DEBUG:
                print(template_group.label, "\tnot found, not required")
            # this element has no matches in the remaining match phrases but is not required so can be skipped
            continue
        elif (not element_sequence or element_sequence["start"] == -1) and group_element.required:
            if DEBUG:
                print(template_group.label, "\tnot found, but required, returning None", group_element.label)
            # this element is required but has no matches, so the phrase matches do not fit the group template
            return None
        elif group_sequence["start"] == group_sequence["end"]:
            # The group is still empty and this element has matches, so set its start and end
            # as the groups start and end sequence
            group_sequence["element_sequences"].append(element_sequence)
            if DEBUG:
                print('adding element sequence:', element_sequence)
                print(template_group.label, "\tfirst found", group_element.label, element_sequence["start"],
                      element_sequence["end"])
                print(template_group.label, "\tcurrent group sequence:", group_sequence["start"], group_sequence["end"])
            group_sequence["start"] = element_sequence["start"]
            group_sequence["end"] = element_sequence["end"]
            if DEBUG:
                print(template_group.label, "\tupdated group sequence:", group_sequence["start"], group_sequence["end"])
            if group_element.required:
                group_sequence["contains_required"] = True
        elif element_sequence["start"] <= group_sequence["end"]:
            # there is no gap between this element's phrase matches and
            # those of the previous elements in the group so update end
            # if this element's matches extend the group end
            group_sequence["element_sequences"].append(element_sequence)
            if DEBUG:
                print('adding element sequence:', element_sequence)
                print(template_group.label, "\tnext found", group_element.label, element_sequence["start"],
                      element_sequence["end"])
                print(template_group.label, "\tcurrent group sequence:", group_sequence["start"], group_sequence["end"])
            if element_sequence["end"] > group_sequence["end"]:
                group_sequence["end"] = element_sequence["end"]
            if DEBUG:
                print(template_group.label, "\tnext sequence:", group_sequence["start"], group_sequence["end"])
        elif group_element.required and element_sequence["start"] > group_sequence["end"]:
            # there is a gap between this element's matches and the previous sequence, but this element is
            # required. Check if parts of the previous sequence are required. If so, the phrase matches do
            # not fit the template group. If not, the previous sequence can be ignored and this required
            # element is the start of the sequence
            if group_sequence["contains_required"]:
                if DEBUG:
                    print(template_group.label, "next required disconnected from previous required")
                return None
            else:
                if DEBUG:
                    print('replacing previous element sequences with current element sequence:', element_sequence)
                group_sequence["element_sequences"] = [element_sequence]
                group_sequence["start"] = element_sequence["start"]
                group_sequence["end"] = element_sequence["end"]
                group_sequence["contains_required"] = True
                if DEBUG:
                    print(template_group.label, "\treset first found", group_element.label, element_sequence["start"],
                          element_sequence["end"])
                    print(template_group.label, "\treset first sequence:", group_sequence["start"], group_sequence["end"])
        else:
            # there are phrase matches in between this element and the previous elements,
            # and this element is not required, so skip this element
            if DEBUG:
                print(template_group.label, "gap between this non-required element and previous sequence. " +
                      "Skipping this element:", group_element.label, element_sequence["start"], element_sequence["end"])
                print('remove element sequence:', element_sequence)
                group_sequence['element_sequences'].pop(-1)
                print('group sequence:', group_sequence['element_sequences'])
            pass
            # return None
    if group_sequence["start"] == -1:
        # none of the group's elements has a phrase match sequence, so the template has no matches
        if DEBUG:
            print(template_group.label, "returning no ordered sequence")
        return None
    if DEBUG:
        print(template_group.label, "returning ordered sequence:", group_sequence["start"], group_sequence["end"])
    return group_sequence


def find_next_unordered_group_match_sequence(phrase_matches: List[PhraseMatch],
                                             template_group: FuzzyTemplateGroupElement,
                                             template_start_index: int) -> Union[None, Dict[str, any]]:
    """Find the next sequence of phrase matches that match an unordered template group element, from a given
    starting point in the list of phrase matches. This function returns None if the template doesn't match.

    :param phrase_matches: a list of phrase matches to be tested against a template element
    :type phrase_matches: List[PhraseMatch]
    :param template_group: a template group element to test the phrase matches against
    :type template_group: FuzzyTemplateGroupElement
    :param template_start_index: the index in the phrase list to start the matching process
    :type template_start_index: int
    :return: a sequence with start and end indexes in the list of phrase matches that match the template group
    :rtype: Union[None, Dict[str, any]]
    """
    # find first phrase match that belongs to first element of group
    group_sequence = initialize_sequence(template_group, -1, -1)
    element_sequences = []
    # print("unordered template group label:", template_group.label, "start_index:", template_start_index)
    for group_element in template_group.elements:
        # print("unordered group element label:", group_element.label)
        if isinstance(group_element, FuzzyTemplateGroupElement):
            # print("checking group element next sequence")
            element_sequence = find_next_group_match_sequence(phrase_matches, group_element, template_start_index)
            # print("returned sequence within unordered template:", group_element.label, element_sequence)
            # if element_sequence:
            #     group_sequence["element_sequences"].append(element_sequence)
        else:
            # print("checking label element next sequence")
            element_start_index = find_next_element_start_index(phrase_matches,
                                                                group_element, template_start_index)
            element_end_index = find_next_element_end_index(phrase_matches,
                                                            group_element, element_start_index)
            element_sequence = initialize_sequence(group_element, element_start_index, element_end_index)
            # print("returned label sequence within unordered template:", group_element.label, element_start_index, element_end_index)
            element_sequence["phrase_matches"] = phrase_matches[element_start_index:element_end_index]
        # group_sequence["element_sequences"].append(element_sequence)
        # element_sequences.append(element_sequence)
        # print(f"unordered group element {group_element.label} indexes:", element_start_index, element_end_index)
        if element_sequence["start"] == -1 and not group_element.required:
            # print("\tnot found, not required")
            # this element has no matches, but is not required
            # so does not contribute to the current template sequence
            continue
        elif element_sequence["start"] == -1 and group_element.required:
            # print("\tnot found, but required")
            # this element has no matches and is required, so the phrase matches
            # do not fit this template group
            return None
        else:
            # this element has matches, so add its sequence to the list of sequences
            # print("\tunordered element sequence", group_element.label, element_start_index, element_end_index)
            # print("\tunordered group sequence:", group_sequence["start"], group_sequence["end"])
            element_sequences.append(element_sequence)
    # print(template_group.label, "sorting element sequences:", len(element_sequences))
    element_sequences.sort(key=lambda x: x["start"])
    for element_sequence in element_sequences:
        # print(template_group.label, "checking sorted element sequence", element_sequence["element_label"],
        #      element_sequence["start"], element_sequence["end"])
        if group_sequence["start"] == -1:
            # this is the first element with matches, so set its sequence as the group sequence
            # print("adding first sequence")
            group_sequence["start"] = element_sequence["start"]
            group_sequence["end"] = element_sequence["end"]
            group_sequence["element_sequences"].append(element_sequence)
        elif element_sequence["start"] <= group_sequence["end"] < element_sequence["end"]:
            # there are no matches in between this element sequence and the previous one,
            # so adjust the end of the group sequence to contain include this element sequence as well
            # print("extending end")
            group_sequence["end"] = element_sequence["end"]
            group_sequence["element_sequences"].append(element_sequence)
        elif element_sequence["end"] <= group_sequence["end"]:
            # this sequences falls within the current group sequence, so add it but leave the group sequence end
            # print("not extending end")
            group_sequence["element_sequences"].append(element_sequence)
        elif element_sequence["start"] > group_sequence["end"] and element_sequence["element"].required:
            # there is a gap between this element's matches and the previous sequence, but this element is
            # required. Check if parts of the previous sequence are required. If so, the phrase matches do
            # not fit the template group. If not, the previous sequence can be ignored and this required
            # element is the start of the sequence
            if group_sequence["contains_required"]:
                # print("next required disconnected from previous required")
                return None
            # the previous sequence has no required element, and this element is required
            # so the matches of the previous elements do not fit this template group
            group_sequence["start"] = element_sequence["start"]
            group_sequence["end"] = element_sequence["end"]
            group_sequence["element_sequences"] = [element_sequence]
        elif element_sequence["start"] > group_sequence["end"]:
            # there are matches in between this element and the previous so this element's
            # matches are part of a next sequence
            break
        else:
            print(template_group.label, "unknown condition")
            print(template_group.label, "group start,end:", group_sequence["start"], group_sequence["end"])
            print(template_group.label, "element start,end:", element_sequence["start"], element_sequence["end"])
            raise ValueError("UNKNOWN CONDITION")
    if group_sequence["start"] == -1:
        # none of the group's elements has a phrase match sequence, so the template has no matches
        # print("returning no unordered sequence:", template_group.label)
        return None
    # print(template_group.label, "returning unordered sequence:", group_sequence["start"], group_sequence["end"])
    return group_sequence


def find_next_group_match_sequence(phrase_matches: List[PhraseMatch],
                                   template_group: FuzzyTemplateGroupElement,
                                   template_start_index: int) -> Union[None, Dict[str, any]]:
    """Find the next sequence of phrase matches that match a template group element, from a given
    starting point in the list of phrase matches. This function returns None if the template doesn't match.

    :param phrase_matches: a list of phrase matches to be tested against a template element
    :type phrase_matches: List[PhraseMatch]
    :param template_group: a template group element to test the phrase matches against
    :type template_group: FuzzyTemplateGroupElement
    :param template_start_index: the index in the phrase list to start the matching process
    :type template_start_index: int
    :return: a sequence with start and end indexes in the list of phrase matches that match the template group
    :rtype: Union[None, Dict[str, any]]
    """
    if template_group.ordered:
        if DEBUG:
            print(template_group.label, "CHECKING FOR FIRST ORDERED SEQUENCE")
        sequence = find_next_ordered_group_match_sequence(phrase_matches, template_group, template_start_index)
        if DEBUG:
            print("returned ordered sequence")
            if sequence:
                print(template_group.label, "\treceived ordered start,end:", sequence["start"], sequence["end"])
            else:
                print(template_group.label, "\treceived no ordered sequence")
        return sequence
    else:
        if DEBUG:
            print(template_group.label, "CHECKING FOR FIRST UNORDERED SEQUENCE")
        sequence = find_next_unordered_group_match_sequence(phrase_matches, template_group, template_start_index)
        if DEBUG:
            if sequence:
                print(template_group.label, "\treceived first start,end:", sequence["start"], sequence["end"])
                print(template_group.label, "CHECKING FOR ADDITIONAL UNORDERED SEQUENCE")
            else:
                print(template_group.label, "no first sequence for")
        while True and sequence:
            next_sequence = find_next_unordered_group_match_sequence(phrase_matches, template_group, sequence["end"])
            if not next_sequence:
                if DEBUG:
                    print(template_group.label, "\treceived no next sequence")
                break
            if next_sequence["start"] != sequence["end"]:
                if DEBUG:
                    print(template_group.label, "\treceived sequence with gap next start,end:", next_sequence["start"], next_sequence["end"])
                break
            elif next_sequence["end"] > sequence["end"]:
                if DEBUG:
                    print(template_group.label, "\treceived sequence with no gap next start,end:", next_sequence["start"], next_sequence["end"])
                sequence["end"] = next_sequence["end"]
                sequence["element_sequences"] += next_sequence["element_sequences"]
            else:
                if DEBUG:
                    print(template_group.label, "\treceived something strange:", next_sequence["start"], next_sequence["end"])
                break
        if DEBUG:
            if sequence:
                print(template_group.label, "RETURN find_next_group_match_sequence:", sequence["start"], sequence["end"])
            else:
                print(template_group.label, "RETURN find_next_group_match_sequence:", sequence)
        return sequence


class FuzzyTemplateSearcher(FuzzyContextSearcher):

    def __init__(self, template: Union[None, FuzzyTemplate] = None, config: Union[None, dict] = None):
        """A fuzzy searcher for finding fuzzy matches in texts and checking if the matches fit a given template.
        The FuzzyTemplateSearcher incorporates a FuzzyContextSearcher for searching phrase matches in texts.
        The phrases are taken from the phrase model that is part of the template. The FuzzyContextSearcher
        uses the default configuration unless a searcher_config is specified that overrides specific properties.

        :param template: a fuzzy template to use for searching
        :type template: FuzzyTemplate
        :param searcher_config: an optional configuration dictionary to configure the FuzzyTemplateSearcher
        """
        super().__init__(config=config)
        self.template: Union[None, FuzzyTemplate] = template if template else None
        self.phrase_model: Union[None, PhraseModel] = template.phrase_model if template else None
        if self.phrase_model:
            self.index_phrase_model(self.phrase_model)

    def __repr__(self):
        return f"{self.__class__.__name__}(template={self.template})"

    def set_template(self, template: FuzzyTemplate) -> None:
        """Set a new template for the searcher and index the corresponding phrase model.

        :param template: a fuzzy template to use for searching
        :type template: FuzzyTemplate
        """
        self.template = template
        self.phrase_model = template.phrase_model
        self.index_phrase_model(self.phrase_model)

    def search_text(self, text: Union[str, Dict[str, str]]) -> List[TemplateMatch]:
        """Search phrases from the registered template's phrase model in the text and check if the resulting matches
        together match the template. This method returns a dictionary including the individual phrase matches and
        any template matches.

        :param text: a text to search in, either as a string or a dictionary with text and an identifier
        :type text: Union[str, Dict[str, str]]
        :return: a dictionary with all phrase matches and template matches
        :rtype: Dict[str, Union[List[PhraseMatch], List[TemplateMatch]]]
        """
        if not self.template:
            raise ValueError("No fuzzy search template registered.")
        phrase_matches = self.find_matches(text)
        template_matches = self.find_template_matches(phrase_matches)
        return template_matches
        # return {"phrase_matches": phrase_matches, "template_matches": template_matches}

    def filter_phrase_matches(self, phrase_matches: List[PhraseMatch]) -> List[PhraseMatch]:
        """Filter a list of phrase matches to only include phrase matches that have at least one label in
        common with the template.

        :param phrase_matches: a list of phrase matches
        :type phrase_matches: List[PhraseMatch]
        :return: a filtered list of phrases matches
        :rtype: List[PhraseMatch]
        """
        # first, check if required elements are present
        if not has_required_matches(phrase_matches, self.template):
            return []
        # second, remove phrase matches that have no label that is part of the template
        return [phrase_match for phrase_match in phrase_matches if self.template.has_label(phrase_match.label)]

    def find_template_matches(self, phrase_matches: List[PhraseMatch]) -> List[TemplateMatch]:
        """Find all the matches that fit a template. The method returns a list of template matches, where
        each template match contains the phrase match that fit the template. There can be multiple
        template matches, if the phrase matches fit a template multiple times.

        :param phrase_matches: a list of phrase matches
        :type phrase_matches: List[PhraseMatch]
        :return: a list of template matches
        :rtype: List[TemplateMatch]
        """
        template_matches: List[TemplateMatch] = []
        # make sure the matches are sorted in order of occurrence in the text
        template_phrase_matches = self.filter_phrase_matches(sorted(phrase_matches, key=lambda x: x.offset))
        sequence_start_index = 0
        # print("num matches:", len(template_phrase_matches))
        # for phrase_match in template_phrase_matches:
        #     print("\t", phrase_match.label, phrase_match.phrase.phrase_string, "\t", phrase_match.string)
        while sequence_start_index < len(template_phrase_matches):
            if DEBUG:
                print("sequence_start_index:", sequence_start_index)
            template_sequence = find_next_group_match_sequence(template_phrase_matches, self.template.root_element,
                                                               sequence_start_index)
            if template_sequence is None:
                break
            else:
                sequence_start_index = template_sequence["end"]
                sequence_matches = phrase_matches[template_sequence["start"]:template_sequence["end"]]
                template_match = TemplateMatch(template=self.template, phrase_matches=sequence_matches,
                                               template_sequence=template_sequence)
                template_matches.append(template_match)
            if DEBUG:
                print("template_sequence:", template_sequence["start"], template_sequence["end"])
            # print("updated sequence_start_index:", sequence_start_index)
            # print("\n\n")
        return template_matches



from __future__ import annotations
from typing import Dict, List, Union
from datetime import datetime
from collections import Counter
import uuid
import string

import fuzzy_search.fuzzy_string as fuzzy_string
from fuzzy_search.fuzzy_string import SkipGram
from fuzzy_search.fuzzy_phrase import Phrase


def validate_match_props(match_phrase: Phrase, match_variant: Phrase,
                         match_string: str, match_offset: int) -> None:
    """Validate match properties.

    :param match_phrase: the phrase that has been matched
    :type match_phrase: Phrase
    :param match_variant: the variant of the phrase that the match is based on
    :type match_variant: Phrase
    :param match_string: the text string that matches the variant phrase
    :type match_string: str
    :param match_offset: the offset of the match string in the text
    :type match_offset: int
    :return: None
    :rtype: None
    """
    if not isinstance(match_phrase, Phrase):
        print(match_phrase)
        print(type(match_phrase))
        raise TypeError('match_phrase MUST be of class Phrase')
    if not isinstance(match_variant, Phrase):
        raise TypeError('match_variant MUST be of class Phrase')
    if not isinstance(match_string, str):
        raise TypeError('match string MUST be a string')
    if len(match_string) == 0:
        print(match_phrase)
        raise ValueError('match string cannot be empty string')
    if not isinstance(match_offset, int):
        raise TypeError('match_offset must be an integer')
    if match_offset < 0:
        raise ValueError('offset cannot be negative')


def adjust_match_start_offset(text: Dict[str, any], match_string: str,
                              match_offset: int) -> Union[int, None]:
    """Adjust the start offset if it is not at a word boundary.

    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param match_string: the candidate match string
    :type match_string: str
    :param match_offset: the text offset of the candidate match string
    :type match_offset: int
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    # adjust the start
    # check if there match initial is a non-word character
    non_word_prefix = fuzzy_string.get_non_word_prefix(match_string)
    if non_word_prefix == "":
        # match does not start with a non-word prefix, so check if it needs to be moved to the left
        if match_offset == 0:
            # match is at the start of text and starts with word characters
            return match_offset
        # if character before match is a word boundary, match offset is good
        if text["text"][match_offset-1:match_offset] in fuzzy_string.non_word_affixes_1:
            return match_offset
        # if penultimate character before match is a word boundary, move offset by -1
        elif match_offset > 1 and text["text"][match_offset-2:match_offset-1] in fuzzy_string.non_word_affixes_1:
            # move match_offset back by 1 to start at word boundary
            return match_offset-1
        # if two characters before match is a word boundary, move offset by -2
        elif match_offset > 2 and text["text"][match_offset-3:match_offset-2] in fuzzy_string.non_word_affixes_1:
            # move match_offset back by 1 to start at word boundary
            return match_offset-2
        # if the three characters preceding match are word characters, the start is wrong
        else:
            return None
    else:
        # match starts with a non-word-prefix, so move offset to after the prefix
        return match_offset + len(non_word_prefix)


def adjust_match_end_offset(phrase_string: str, candidate_string: str,
                            text: Dict[str, any], end_offset: int, punctuation: str) -> Union[int, None]:
    """Adjust the end offset if it is not at a word boundary.

    :param phrase_string: the phrase string
    :type phrase_string: str
    :param candidate_string: the candidate match string
    :type candidate_string: str
    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param end_offset: the text offset of the candidate match string
    :type end_offset: int
    :param punctuation: the set of characters to treat as punctuation
    :type punctuation: str
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    # ugly hack: if phrase string ends with punctuation, use only whitespace as word end boundary
    whitespace_only = True if phrase_string[-1] in punctuation else False
    phrase_end = map_string(phrase_string[-3:], punctuation)
    match_end = map_string(candidate_string[-3:], punctuation, whitespace_only=whitespace_only)
    text_suffix = map_string(text["text"][end_offset:end_offset+3], punctuation, whitespace_only=whitespace_only)
    # print(f"match_end: {candidate_string[-3:]: <4}\ttext_suffix: {text['text'][end_offset:end_offset+3]: >4}")
    # print(f"mapped suffixes - match_end: #{match_end}#\ttext_suffix: #{text_suffix}#")
    return calculate_end_shift(phrase_end, match_end, text_suffix, end_offset)


def adjust_match_offsets(phrase_string: str, candidate_string: str,
                         text: Dict[str, any], candidate_start_offset: int,
                         candidate_end_offset: int,
                         punctuation: str = string.punctuation) -> Union[Dict[str, Union[str, int]], None]:
    """Adjust the end offset if it is not at a word boundary.

    :param phrase_string: the phrase string
    :type phrase_string: str
    :param candidate_string: the candidate match string
    :type candidate_string: str
    :param text: the text object that contains the candidate match string
    :type text: Dict[str, any]
    :param candidate_start_offset: the text offset of the start of the candidate match string
    :type candidate_start_offset: int
    :param candidate_end_offset: the text offset of the end of the candidate match string
    :type candidate_end_offset: int
    :param punctuation: the set of characters to treat as punctuation (defaults to string.punctuation)
    :type punctuation: str
    :return: the adjusted offset or None if the required adjustment is too big
    :rtype: Union[int, None]
    """
    # print("phrase string:", phrase_string)
    # print("adjusting candidate string:", candidate_string)
    if punctuation is None:
        punctuation = string.punctuation
    # print("candidate_start_offset:", candidate_start_offset)
    match_start_offset = adjust_match_start_offset(text, candidate_string, candidate_start_offset)
    # print("match_start_offset:", match_start_offset)
    if match_start_offset is None:
        return None
    match_end_offset = adjust_match_end_offset(phrase_string, candidate_string,
                                               text, candidate_end_offset, punctuation)
    if match_end_offset is None:
        return None
    elif match_end_offset <= match_start_offset:
        return None
    return {
        "match_string": text["text"][match_start_offset:match_end_offset],
        "match_start_offset": match_start_offset,
        "match_end_offset": match_end_offset
    }


def map_string(affix_string: str, punctuation: str, whitespace_only: bool = False) -> str:
    """Turn affix string into type char representation. Types are 'w' for non-whitespace char,
    and 's' for whitespace char.

    :param affix_string: a string
    :type: str
    :param punctuation: the set of characters to treat as punctuation
    :type punctuation: str
    :param whitespace_only: whether to treat only whitespace as word boundary or also include (some) punctuation
    :type whitespace_only: bool
    :return: the type char representation
    :rtype: str
    """
    if whitespace_only:
        return "".join(["s" if char == " " else "w" for char in affix_string])
    else:
        return "".join(["s" if char == " " or char in punctuation else "w" for char in affix_string])


def calculate_end_shift(phrase_end: str, match_end: str, text_suffix: str, end_offset: int):
    if phrase_end == match_end:
        if text_suffix == "" or text_suffix.startswith("s"):
            return end_offset
    if phrase_end.endswith("s") and match_end.endswith("s"):
        return end_offset
    if match_end == "wss":
        return end_offset - 2
    if phrase_end == "www":
        if match_end == "www":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            elif text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 2
            elif text_suffix.startswith("www"):
                return None
        if match_end == "wws":
            return end_offset - 1
        if match_end == "wsw":
            if text_suffix == "" or text_suffix.startswith("s"):
                # we assume the whitespace in the match is a misrecognised word character
                return end_offset
            if text_suffix.startswith("w"):
                # we assume the whitespace in the match is correct
                return end_offset - 2
        if match_end == "sww":
            if text_suffix == "" or text_suffix.startswith("s"):
                # we assume the whitespace in the match is a misrecognised word character
                return end_offset
            elif text_suffix.startswith("w"):
                # we assume the whitespace in the match is correct
                return None
        if match_end == "sws":
            # we assume the first whitespace in the match is a misrecognised word character
            return end_offset - 1
        if match_end == "ssw":
            return None
        else:
            return None
    if phrase_end == "wws":
        if match_end == "www":
            if text_suffix == "":
                return end_offset
            elif text_suffix.startswith("s"):
                return end_offset + 1
            elif text_suffix == "w":
                return None
            elif text_suffix.startswith("ws"):
                return end_offset + 2
            elif text_suffix.startswith("ww"):
                return None
            else:
                return None
        elif match_end.startswith("ws"):
            return end_offset - 1
        elif match_end.startswith("s"):
            return end_offset - 2
        else:
            return None
    if phrase_end == "sww":
        if match_end == "sww":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            else:
                return None
        elif match_end == "sws":
            return end_offset - 1
        elif match_end == "www":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        elif match_end == "wsw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            if text_suffix == "ws" or text_suffix.startswith("ws"):
                return end_offset + 1
            if text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 2
            else:
                return None
        elif match_end == "ssw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            elif text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            elif text_suffix == "ww" or text_suffix.startswith("wws"):
                return end_offset + 1
            else:
                return None
        else:
            return None
    if phrase_end == "sws":
        if match_end == "www":
            if text_suffix == "sw" or text_suffix == "sws":
                return end_offset + 2
            else:
                return None
        elif match_end == "sww":
            return end_offset - 2
        elif match_end == "wsw":
            if text_suffix == "":
                return end_offset
            if text_suffix.startswith("s"):
                return end_offset + 1
            else:
                return end_offset - 1
        else:
            return None
    if phrase_end == "wsw":
        if match_end == "wsw":
            if text_suffix == "w" or text_suffix.startswith("ws"):
                return end_offset + 1
            else:
                return None
        if match_end == "www":
            if text_suffix == "" or text_suffix == "s":
                return end_offset
            elif text_suffix.startswith("w"):
                return None
            elif text_suffix == "sw" or text_suffix == "sws":
                return end_offset + 2
            else:
                return None
        if match_end == "sww":
            if text_suffix == "":
                return end_offset + 1
            elif text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        if match_end == "ssw":
            if text_suffix == "" or text_suffix.startswith("s"):
                return end_offset
            else:
                return None
        else:
            return None
    if len(phrase_end) < 3:
        if phrase_end == match_end:
            return end_offset
        else:
            return None
    else:
        details = f"phrase_end {phrase_end}, match_end {match_end}, text_suffix {text_suffix}"
        raise ValueError(f"combination not captured: {details}")


###################
# Candidate class #
###################

class Candidate:

    def __init__(self, phrase: Phrase, max_length_variance: int = 1):
        """Create a Candidate instance for a given Phrase object.

        :param phrase: a phrase object
        :type phrase: Phrase
        """
        self.skipgram_set = set()
        self.skipgram_list: List[SkipGram] = []
        self.skipgram_count = Counter()
        self.phrase = phrase
        self.max_length_variance = max_length_variance
        self.max_length = len(self.phrase.phrase_string) + self.max_length_variance
        self.match_start_offset: int = -1
        self.match_end_offset: int = -1
        self.match_string: Union[None, str] = None
        self.skipgram_overlap: float = 0.0

    def __repr__(self):
        return f'Candidate(' + \
               f'phrase: "{self.phrase.phrase_string}", match_string: "{self.match_string}", ' + \
               f'match_start_offset: {self.match_start_offset}, match_end_offset: {self.match_end_offset})'

    def add_skip_match(self, skipgram: SkipGram) -> None:
        """Add a skipgram match between a text and a phrase ot the candidate.

        :param skipgram: a matching skipgram
        :type skipgram: SkipGram
        """
        if len(self.skipgram_list) == 0 and skipgram.string not in self.phrase.early_skipgram_index:
            # print("skipping skipgram as first for candidate:", skipgram.string)
            return None
        self.skipgram_set.add(skipgram.string)
        self.skipgram_list.append(skipgram)
        if self.match_start_offset is None or self.match_start_offset < 0:
            self.match_start_offset = self.get_match_start_offset()
        if skipgram.offset + skipgram.length > self.match_end_offset:
            self.match_end_offset = skipgram.offset + skipgram.length
        self.skipgram_count.update([skipgram.string])
        # print("\tadd - skipgram:", skipgram.string, skipgram.offset)
        # print("\tadd - match length:", self.skip_match_length())
        # print("\tadd - list:", [skip.string for skip in self.skipgram_list])
        # check if the candidate string is too long to match the phrase
        # if too long, remove the first skipgrams until the string is short enough
        while self.skip_match_length() > self.max_length and len(self.skipgram_list) > 0:
            self.remove_first_skip()
            self.match_start_offset = self.get_match_start_offset()
            # print("\tremove - too long - length:", self.skip_match_length())
            # print("\tremove - too long - list:", [skip.string for skip in self.skipgram_list])
            # print("\tremove - too long - start:", self.match_start_offset, "\tend:", self.match_end_offset)
        while len(self.skipgram_list) > 0 and self.skipgram_list[0].string not in self.phrase.early_skipgram_index:
            self.remove_first_skip()
            self.match_start_offset = self.get_match_start_offset()
            # print("\tremove - no start - length:", self.skip_match_length())
            # print("\tremove - no start - list:", [skip.string for skip in self.skipgram_list])
            # print("\tremove - no start - start:", self.match_start_offset, "\tend:", self.match_end_offset)

    def shift_start_skip(self) -> bool:
        """Check if there is a later skip that is a better start."""
        if self.skip_match_length() <= len(self.phrase.phrase_string):
            return False
        start_skip = self.skipgram_list[0]
        start_phrase_offset = self.phrase.skipgram_index[start_skip.string][0].offset
        best_start_phrase_offset = start_phrase_offset
        best_start_index = 0
        best_start_skip = start_skip
        for si, skip in enumerate(self.skipgram_list):
            skip_phrase_offset = self.phrase.skipgram_index[skip.string][0].offset
            if skip.offset - start_skip.offset > self.skip_match_length() - len(self.phrase.phrase_string):
                # stop looking for better start when remaining skips result in too short match length
                break
            if skip.offset > best_start_skip.offset and skip_phrase_offset <= best_start_phrase_offset:
                best_start_index = si
                best_start_skip = skip
                best_start_phrase_offset = skip_phrase_offset
            if skip.string not in self.phrase.early_skipgram_index:
                break
        for _ in range(0, best_start_index):
            self.remove_first_skip()
        self.match_start_offset = self.get_match_start_offset()
        return best_start_index > 0

    def remove_first_skip(self) -> None:
        """Remove the first matching skipgram from the list and update the count and set."""
        first_skip = self.skipgram_list.pop(0)
        # print('removing first skip')
        # reduce count of first skipgram by 1
        self.skipgram_count[first_skip.string] -= 1
        # if count has dropped to zero, remove skipgram from the set
        if self.skipgram_count[first_skip.string] == 0:
            self.skipgram_set.remove(first_skip.string)

    def skip_match_length(self) -> int:
        """Return the length of the matching string.

        :return: difference between start and end offset
        :rtype: int
        """
        if self.match_start_offset is None:
            return 0
        return self.match_end_offset - self.match_start_offset

    def is_match(self, skipgram_threshold: float):
        """Check if the candidate is a likely match for its corresponding phrase.

        :param skipgram_threshold: the threshold to for how many skipgrams have to match between candidate and phrase
        :type skipgram_threshold: float
        :return: a boolean whether this candidate is a likely match for the phrase
        :rtype: bool
        """
        if len(self.skipgram_list) == 0:
            # there are no matching skipgrams, so no matching string
            # print('NO MATCH: there are no matching skipgrams, so no matching string')
            return False
        if self.skipgram_list[0].string not in self.phrase.early_skipgram_index:
            # the first skipgram of candidate is not in the early skipgrams of phrase
            # print('NO MATCH: the first skipgram of candidate is not in the early skipgrams of phrase')
            return False
        # length of current skip matches should be no longer than phrase plus max length variance
        # but also no shorter than phrase minus max length variance minus late skips offset from end
        # of phrase.
        if self.skip_match_length() > len(self.phrase.phrase_string) + self.max_length_variance:
            # print('NO MATCH: skip match length is longer than max length variance of phrase')
            return False
        elif self.skip_match_length() < self.phrase.late_threshold - self.max_length_variance:
            # print('phrase length:', len(self.phrase.phrase_string))
            # print('skip match length:', self.skip_match_length())
            # print('max length variance:', self.max_length_variance)
            # print('late threshold:', self.phrase.late_threshold)
            # print('NO MATCH: skip match length is not with max length variance of phrase')
            return False
        # print(self.get_skip_set_overlap())
        # print('late threshold:', self.phrase.late_threshold)
        if self.skipgram_list[-1].string not in self.phrase.late_skipgram_index:
            # print("NO MATCH: last skip not in late index")
            # the last skipgram of candidate is not in the late skipgrams of phrase
            return False
        if self.get_skip_set_overlap() < skipgram_threshold:
            # print("below skipgram threshold:", self.get_skip_set_overlap(), skipgram_threshold)
            return False
        else:
            return True

    def get_skip_set_overlap(self) -> float:
        """Calculate and set skipgram overlap between text and phrase skipgram matches.

        :return: the skipgram overlap
        :rtype: float
        """
        self.skipgram_overlap = len(self.skipgram_set) / len(self.phrase.skipgram_set)
        return self.skipgram_overlap

    def get_skip_count_overlap(self) -> float:
        """Calculate deviation of candidate skipgrams from phrase skipgrams.

        :return: the skipgram overlap (-inf, 1.0]
        :rtype: float
        """
        diff = 0
        total = 0
        for skipgram_string, count in self.skipgram_count.items():
            diff += abs(count - self.phrase.skipgram_freq[skipgram_string])
            total += count
        return (total - diff) / self.phrase.num_skipgrams

    def get_match_start_offset(self) -> Union[None, int]:
        """Calculate the start offset of the match.

        :return: the start offset of the match
        :rtype: int
        """
        if len(self.skipgram_list) == 0:
            return None
        first_skip = self.skipgram_list[0]
        first_skip_in_phrase = self.phrase.skipgram_index[first_skip.string][0]
        match_start_offset = self.skipgram_list[0].offset - first_skip_in_phrase.offset
        # print("in match:", first_skip.string, first_skip.offset)
        # print("in phrase:", first_skip_in_phrase.string, first_skip_in_phrase.offset)
        # print("match_start_offset:", match_start_offset)
        return 0 if match_start_offset < 0 else match_start_offset

    def get_match_string(self, text: Dict[str, any]) -> Union[str, None]:
        """Find the matching string of a candidate fuzzy match between a text and a phrase.

        :param text: the text object from which the candidate was derived
        :type text: Dict[str, any]
        :return: the matching string
        :rtype: str
        """
        if self.match_start_offset == self.match_end_offset:
            raise ValueError('start and end offset cannot be the same')
        if self.match_start_offset > self.match_end_offset:
            raise ValueError('start offset cannot be bigger than end offset')
        return text["text"][self.match_start_offset:self.match_end_offset]
    # TODO: check if first to last offset is too long
    # if not, the match string is probably fine
    # if it is, find the best substring

    def same_candidate(self, other: Candidate):
        """Check if this candidate has the same start and end offsets as another candidate.

        :param other: another candidate for the same phrase and text.
        :type other: Candidate
        :return: this candidate match has the same offsets as the other candidate
        :rtype: bool
        """
        if self.match_start_offset != other.match_start_offset:
            return False
        if self.match_end_offset != other.match_end_offset:
            return False
        else:
            return True


###############
# Match class #
###############

class PhraseMatch:

    def __init__(self, match_phrase: Phrase, match_variant: Phrase, match_string: str,
                 match_offset: int, text_id: Union[None, str] = None,
                 match_scores: dict = None, match_label: Union[str, List[str]] = None,
                 match_id: str = None):
        # print("Match class match_phrase:", match_phrase)
        validate_match_props(match_phrase, match_variant, match_string, match_offset)
        self.id = match_id if match_id else str(uuid.uuid4())
        self.phrase = match_phrase
        self.label = match_phrase.label
        if match_label:
            self.label = match_label
        self.metadata = {}
        self.variant = match_variant
        self.string = match_string
        self.offset = match_offset
        self.end = self.offset + len(self.string)
        self.text_id = text_id
        self.character_overlap: Union[None, float] = None
        self.ngram_overlap: Union[None, float] = None
        self.skipgram_overlap: Union[None, float] = None
        self.levenshtein_similarity: Union[None, float] = None
        if match_scores:
            self.character_overlap = match_scores['char_match']
            self.ngram_overlap = match_scores['ngram_match']
            self.levenshtein_similarity = match_scores['levenshtein_similarity']
        self.created = datetime.now()

    def __repr__(self):
        return f'PhraseMatch(' + \
            f'phrase: "{self.phrase.phrase_string}", variant: "{self.variant.phrase_string}",' + \
            f'string: "{self.string}", offset: {self.offset})'

    @property
    def label_list(self) -> List[str]:
        if isinstance(self.label, str):
            return [self.label]
        elif isinstance(self.label, list):
            return self.label
        else:
            return []

    def has_label(self, label: str):
        if isinstance(self.label, str):
            return label == self.label
        else:
            return label in self.label

    def json(self) -> dict:
        data = {
            "type": "PhraseMatch",
            "phrase": self.phrase.phrase_string,
            "variant": self.variant.phrase_string,
            "string": self.string,
            "offset": self.offset,
            "label": self.label,
            "text_id": self.text_id,
            "match_scores": {
                "char_match": self.character_overlap,
                "ngram_match": self.ngram_overlap,
                "levenshtein_similarity": self.levenshtein_similarity
            }
        }
        if "label" in self.phrase.metadata:
            data["label"] = self.phrase.metadata["label"]
        return data

    def add_scores(self, skipgram_overlap: Union[None, float] = None) -> None:
        """Compute overlap and similarity scores between the match variant and the match string
        and add these to the match object.

        :param skipgram_overlap: the overlap in skipgrams between match string and match variant
        :type skipgram_overlap: Union[float, None]
        :return: None
        :rtype: None
        """
        self.character_overlap = self.score_character_overlap()
        self.ngram_overlap = self.score_ngram_overlap()
        self.levenshtein_similarity = self.score_levenshtein_similarity()
        if skipgram_overlap is not None:
            self.skipgram_overlap = skipgram_overlap

    def score_character_overlap(self):
        """Return the character overlap between the variant phrase_string and the match_string

        :return: the character overlap as proportion of the variant phrase string
        :rtype: float
        """
        if not self.character_overlap:
            self.character_overlap = fuzzy_string.score_char_overlap_ratio(self.variant.phrase_string, self.string)
        return self.character_overlap

    def score_ngram_overlap(self) -> float:
        """Return the ngram overlap between the variant phrase_string and the match_string

        :return: the ngram overlap as proportion of the variant phrase string
        :rtype: float
        """
        if not self.ngram_overlap:
            self.ngram_overlap = fuzzy_string.score_ngram_overlap_ratio(self.variant.phrase_string,
                                                                        self.string, self.phrase.ngram_size)
        return self.ngram_overlap

    def score_levenshtein_similarity(self):
        """Return the levenshtein similarity between the variant phrase_string and the match_string

        :return: the levenshtein similarity as proportion of the variant phrase string
        :rtype: float
        """
        if not self.levenshtein_similarity:
            self.levenshtein_similarity = fuzzy_string.score_levenshtein_similarity_ratio(self.variant.phrase_string,
                                                                                          self.string)
        return self.levenshtein_similarity

    def overlaps(self, other: PhraseMatch) -> bool:
        """Check if the match string of this match object overlaps with the match string of another match object.

        :param other: another match object
        :type other: PhraseMatch
        :return: a boolean indicating whether the match_strings of the two objects overlap in the source text
        :rtype: bool"""
        if self.text_id is not None and self.text_id != other.text_id:
            return False
        if self.offset <= other.offset < self.end:
            return True
        elif other.offset <= self.offset < other.end:
            return True
        else:
            return False

    def as_web_anno(self) -> Dict[str, any]:
        """Turn match object into a W3C Web Annotation representation"""
        if not self.text_id:
            raise ValueError('Cannot make target: match object has no text_id')
        body_match = [
            {
                'type': 'TextualBody',
                'purpose': 'tagging',
                'format': 'text',
                'value': self.phrase.phrase_string
            },
            {
                'type': 'TextualBody',
                'purpose': 'highlighting',
                'format': 'text',
                'value': self.string
            }
        ]
        if self.variant.phrase_string != self.string:
            correction = {
                'type': 'TextualBody',
                'purpose': 'correcting',
                'format': 'text',
                'value': self.variant.phrase_string
            }
            body_match.append(correction)
        if self.label:
            classification = {
                'type': 'TextualBody',
                'purpose': 'classifying',
                'format': 'text',
                'value': self.label
            }
            body_match.append(classification)
        return {
            "@context": "http://www.w3.org/ns/anno.jsonld",
            "id": self.id,
            "type": "Annotation",
            "motivation": "classifying",
            "created": self.created.isoformat(),
            "generator": {
                "id": "https://github.com/marijnkoolen/fuzzy-search",
                "type": "Software",
                "name": f"fuzzy-search"
            },
            "target": {
                "source": self.text_id,
                "selector": {
                    "type": "TextPositionSelector",
                    "start": self.offset,
                    "end": self.end
                }
            },
            "body": body_match
        }


class PhraseMatchInContext(PhraseMatch):

    def __init__(self, match: PhraseMatch, text: Union[str, dict] = None, context: str = None,
                 context_start: int = None, context_end: int = None,
                 prefix_size: int = 20, suffix_size: int = 20):
        super().__init__(match_phrase=match.phrase, match_variant=match.variant, match_string=match.string,
                         match_offset=match.offset, text_id=match.text_id)
        """MatchInContext extends a Match object with surrounding context from the text document that the match
        phrase was taken from. Alternatively, the context can be submitted.

        :param text: the text (string or dictionary with 'text' and 'id' properties) that the match phrase was taken from
        :type text: Union[str, dict]
        :param context: the context string around the match phrase 
        :type context: Union[str, dict]
        :param match: the match phrase object
        :type match: Match
        :param context_start: the start offset of the context in the original text
        :type context_start: int
        :param context_end: the end offset of the context in the original text
        :type context_end: int
        :param prefix_size: the size of the prefix window
        :type prefix_size: int
        :param suffix_size: the size of the suffix window
        :type suffix_size: int 
        """
        self.character_overlap = match.character_overlap
        self.ngram_overlap = match.ngram_overlap
        self.levenshtein_similarity = match.levenshtein_similarity
        self.prefix_size = prefix_size
        self.suffix_size = suffix_size
        if text:
            if isinstance(text, str):
                text = {"text": text, "id": match.text_id}
            self.context_start = match.offset - prefix_size if match.offset >= prefix_size else 0
            self.context_end = match.end + suffix_size if len(text["text"]) > match.end + suffix_size else len(text["text"])
            self.context = text["text"][self.context_start:self.context_end]
        elif context:
            self.context = context
            self.context_start = context_start
            self.context_end = context_end
        self.prefix = text["text"][self.context_start:match.offset]
        self.suffix = text["text"][match.end:self.context_end]

    def __repr__(self):
        return f'PhraseMatchInContext(' + \
               f'phrase: "{self.phrase.phrase_string}", variant: "{self.variant.phrase_string}",' + \
               f'string: "{self.string}", offset: {self.offset}), context: "{self.context}"'

    def json(self):
        json_data = super().json()
        json_data["context_start"] = self.context_start
        json_data["context_end"] = self.context_end
        json_data["context"] = self.context
        json_data['prefix_size'] = self.prefix_size
        json_data['suffix_size'] = self.suffix_size
        json_data['prefix'] = self.prefix
        json_data['suffix'] = self.suffix
        return json_data

    def as_web_anno(self) -> Dict[str, any]:
        match_anno = super().as_web_anno()
        position_selector = match_anno['target']['selector']
        quote_selector = {
            'type': 'TextQuoteSelector',
            'prefix': self.prefix,
            'exact': self.string,
            'suffix': self.suffix
        }
        match_anno['target']['selector'] = [position_selector, quote_selector]
        return match_anno


def phrase_match_from_json(match_json: dict) -> PhraseMatch:
    match_phrase = Phrase(match_json['phrase'])
    match_variant = Phrase(match_json['variant'])
    phrase_match = PhraseMatch(match_phrase, match_variant, match_json['string'],
                               match_offset=match_json['offset'],
                               match_scores=match_json['match_scores'],
                               match_label=match_json['label'])
    if 'context' in match_json:
        phrase_match = PhraseMatchInContext(phrase_match, context=match_json['context'],
                                            prefix_size=match_json['prefix_size'],
                                            suffix_size=match_json['suffix_size'],
                                            context_start=match_json['context_start'],
                                            context_end=match_json['context_end'])
    return phrase_match



default_config = {
    'char_match_threshold': 0.5,
    'ngram_threshold': 0.5,
    'skipgram_threshold': 0.3,
    'levenshtein_threshold': 0.5,
    'include_variants': False,
    'filter_distractors': False
}


from unittest import TestCase

from fuzzy_search.fuzzy_phrase_model import PhraseModel
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_template import FuzzyTemplate
from fuzzy_search.fuzzy_template_searcher import FuzzyTemplateSearcher, get_phrase_match_list_labels
from fuzzy_search.fuzzy_match import PhraseMatch

from data.demo_data import DemoData


class TestFuzzyTemplateSearcher(TestCase):

    def setUp(self) -> None:
        self.phrase_model = PhraseModel([{"phrase": "test", "label": "test_label"}])
        self.template = FuzzyTemplate(self.phrase_model, template_json=["test_label"])

    def test_can_make_searcher(self):
        searcher = FuzzyTemplateSearcher()
        self.assertEqual(searcher.template, None)

    def test_can_add_template_at_init(self):
        searcher = FuzzyTemplateSearcher(template=self.template)
        self.assertEqual(searcher.template, self.template)

    def test_can_add_template_later(self):
        searcher = FuzzyTemplateSearcher()
        searcher.set_template(self.template)
        self.assertEqual(searcher.template, self.template)

    def test_add_template_sets_phrase_model(self):
        searcher = FuzzyTemplateSearcher()
        searcher.set_template(self.template)
        self.assertEqual(searcher.phrase_model, self.phrase_model)

    def test_configure_ngram_size(self):
        searcher = FuzzyTemplateSearcher(config={"ngram_size": 3})
        phrase_model = PhraseModel([{"phrase": "test", "label": "test_label"}], config={"ngram_size": 3})
        template = FuzzyTemplate(phrase_model, template_json=["test_label"])
        searcher.set_template(template)
        self.assertEqual(searcher.ngram_size, 3)

    def test_throws_error_for_mismatch_ngram_size(self):
        searcher = FuzzyTemplateSearcher(config={"ngram_size": 3})
        phrase_model = PhraseModel([{"phrase": "test", "label": "test_label"}], config={"ngram_size": 2})
        template = FuzzyTemplate(phrase_model, template_json=["test_label"])
        error = None
        try:
            searcher.set_template(template)
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_can_search_text(self):
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.search_text({"text": "this is a test", "id": "urn:this:text"})
        self.assertEqual(len(template_matches), 1)

    def test_search_text_returns_template_matches(self):
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.search_text({"text": "this is a test", "id": "urn:this:text"})
        self.assertEqual(len(template_matches), 1)


class TestFuzzyTemplateSearcherWithRealData(TestCase):

    def setUp(self) -> None:
        demo_data = DemoData()
        self.auction_data = demo_data.get_dataset("auction_advertisements")
        self.phrase_model = PhraseModel(model=self.auction_data["phrases"])
        self.template = FuzzyTemplate(phrase_model=self.phrase_model, template_json=self.auction_data["template"],
                                      ignore_unknown=True)
        self.auction_tests = self.auction_data["tests"]

    def prep_test(self, test_name):
        test_data = self.auction_tests[test_name]
        matches = []
        for index, match_label in enumerate(test_data["match_sequence"]):
            match_phrase = None
            for phrase in self.phrase_model.get_phrases():
                if phrase.has_label(match_label):
                    match_phrase = phrase
                    break
            if match_phrase is None:
                raise ValueError(f"No phrase with label {match_label}")
            match = PhraseMatch(match_phrase=match_phrase, match_variant=match_phrase,
                                match_string=match_label, match_offset=index)
            match.label = match_label
            matches.append(match)
        return matches, test_data

    def test_search_text_finds_template_with_auction_test_1(self):
        test_matches, test_data = self.prep_test("test1")
        test_template_labels = test_data["template_matches"][0]
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.find_template_matches(test_matches)
        self.assertEqual(len(template_matches), 1)
        labels = get_phrase_match_list_labels(template_matches[0].phrase_matches)
        self.assertEqual(len(labels), len(test_template_labels))
        is_same = True
        for li, label in enumerate(labels):
            if label != test_template_labels[li]:
                is_same = False
                break
        self.assertEqual(is_same, True)

    def test_search_text_finds_template_with_auction_test_2(self):
        test_matches, test_data = self.prep_test("test2")
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.find_template_matches(test_matches)
        self.assertEqual(len(template_matches), test_data["num_template_matches"])

    def test_search_text_finds_template_with_auction_test_3(self):
        test_matches, test_data = self.prep_test("test3")
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.find_template_matches(test_matches)
        self.assertEqual(len(template_matches), test_data["num_template_matches"])

    def test_search_text_finds_template_with_auction_test_4(self):
        test_matches, test_data = self.prep_test("test4")
        searcher = FuzzyTemplateSearcher(template=self.template)
        template_matches = searcher.find_template_matches(test_matches)
        self.assertEqual(len(template_matches), test_data["num_template_matches"])



from unittest import TestCase

from fuzzy_search.fuzzy_match import PhraseMatch, PhraseMatchInContext
from fuzzy_search.fuzzy_context_searcher import FuzzyContextSearcher
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_phrase_model import PhraseModel


class TestFuzzyContextSearcher(TestCase):

    def setUp(self) -> None:
        self.text = "This string contains test text."
        self.phrase = Phrase("test")
        self.searcher = FuzzyContextSearcher()
        self.match = PhraseMatch(self.phrase, self.phrase, "test", 21)

    def test_fuzzy_context_searcher(self):
        error = None
        try:
            FuzzyContextSearcher()
        except any as err:
            error = err
        self.assertEqual(error, None)

    def test_fuzzy_context_searcher_can_add_context(self):
        match_in_context = self.searcher.add_match_context(self.match, self.text)
        self.assertEqual(match_in_context.context_start, 0)

    def test_fuzzy_context_searcher_finds_match_with_context(self):
        self.searcher.index_phrases(["test"])
        matches = self.searcher.find_matches(self.text)
        self.assertEqual(len(matches[0].context) > 0, True)

    def test_fuzzy_context_searcher_can_set_context_size(self):
        self.searcher.index_phrases(["test"])
        matches = self.searcher.find_matches(self.text, prefix_size=0, suffix_size=10)
        self.assertEqual(len(matches[0].prefix) == 0, True)
        self.assertEqual(len(matches[0].suffix) > 0, True)

    def test_fuzzy_context_searcher_can_search_context(self):
        match_in_context = self.searcher.add_match_context(self.match, self.text)
        phrase_model = PhraseModel(model=[{"phrase": "contains"}])
        context_searcher = FuzzyContextSearcher()
        context_searcher.index_phrase_model(phrase_model)
        matches = context_searcher.find_matches_in_context(match_in_context)
        self.assertEqual(len(matches), 1)



from __future__ import annotations
from typing import Dict, List, Set, Union

from fuzzy_search.fuzzy_phrase_model import PhraseModel
from fuzzy_search.fuzzy_phrase import Phrase


def validate_element_properties(label: str, required: bool = False, cardinality: str = "multi",
                                next_label: Union[None, str, List[str]] = None,
                                next_distance_max: Union[None, int] = None,
                                variable: bool = False) -> None:
    """Validate the properties of a FuzzyTemplate element.

    :param label: the label of the element, which can be a single string or a list of strings
    :type label: Union[str, List[str]]
    :param required: whether or not the element must match for the template to match
    :type required: bool
    :param cardinality: whether the element can occur only once (default) or multiple times in a template match.
    :type cardinality: str
    :param next_label: what the label of the next element should be. Use a list of labels for multiple options.
    :type next_label: Union[str, List[str]]
    :param next_distance_max: the maximum distance allowed between this element and the next element in the template
    :type next_distance_max: int
    :param variable: flag to indicate the element has no phrases but has variable text (default is False)
    :type variable: bool
    """
    if not isinstance(label, str):
        raise ValueError("label must be string")
    if required is None:
        pass
    elif not isinstance(required, bool):
        raise ValueError("'required' property must be a boolean value")
    if cardinality is None:
        pass
    elif not isinstance(cardinality, str) or cardinality not in ["single", "multi"]:
        raise ValueError("cardinality must be a string with either 'single' or 'multi' as value")
    if next_label is None:
        pass
    elif isinstance(next_label, list):
        for label in next_label:
            if not isinstance(label, str):
                raise ValueError("next_label list items must be strings")
    elif not isinstance(next_label, str):
        raise ValueError("next_label must be string or list of strings")
    if next_distance_max is None:
        pass
    elif not isinstance(next_distance_max, int) or next_distance_max < 0:
        raise ValueError("next_distance_max must be an positive integer")
    if not isinstance(variable, bool):
        raise ValueError("'variable' flag must be a boolean")


class FuzzyTemplateElement:

    def __init__(self, label: Union[None, str, List[str]], element_type: str, required: bool):
        self.label = label
        self.type = element_type
        self.required = required


class FuzzyTemplateLabelElement(FuzzyTemplateElement):

    def __init__(self, label: str, required: bool = False, cardinality: str = "single",
                 next_label: Union[None, str, List[str]] = None, next_distance_max: Union[None, int] = None,
                 variable: bool = False):
        """A FuzzyTemplate element with properties to define its role in the template it is part of.
        The 'label' property is the only required property. All elements must have either a single string or
        list of strings as label.

        :param label: the label of the element, which can be a single string or a list of strings
        :type label: Union[str, List[str]]
        :param required: whether or not the element must match for the template to match
        :type required: bool
        :param cardinality: whether the element can occur only once or multiple times in a template match. The default
        value is 'multi'
        :type cardinality: str
        :param next_label: what the label of the next template element should be. This can be a list of labels to
        allow different types of element to come next
        :type next_label: Union[str, List[str]]
        :param next_distance_max: the maximum distance allowed between this element and the next element in the template
        :type next_distance_max: int
        :param variable: flag to indicate the element has no phrases but has variable text (default is False)
        :type variable: bool
        """
        validate_element_properties(label, required, cardinality, next_label, next_distance_max)
        super().__init__(label, "label", required)
        self.cardinality = cardinality if cardinality is not None else "single"
        self.next_label = next_label
        self.next_distance_max = next_distance_max
        self.variable = variable

    def __repr__(self):
        return f"FuzzyTemplateElement(label='{self.label}', required={self.required}, cardinality='{self.cardinality}'"


class FuzzyTemplateGroupElement(FuzzyTemplateElement):

    def __init__(self, elements: List[FuzzyTemplateElement],
                 label: Union[str, None] = None, ordered: bool = True, required: bool = False):
        super().__init__(label, "group", required)
        self.ordered = ordered
        self.elements = elements
        self.group_element_labels: Set[str] = set()
        self.has_variable_element = False
        for element in self.elements:
            if isinstance(element, FuzzyTemplateGroupElement):
                self.group_element_labels = self.group_element_labels.union(element.group_element_labels)
                self.has_variable_element = element.has_variable_element
            else:
                self.group_element_labels.add(element.label)
                if isinstance(element, FuzzyTemplateLabelElement) and element.variable:
                    self.has_variable_element = True
            if element.required:
                # override the received required value if one of its sub-elements is required
                self.required = True

    def __repr__(self):
        return f"FuzzyTemplateGroup(label='{self.label}', required={self.required}, ordered='{self.ordered}'"


def generate_label_from_json(label: str, element_info: dict) -> FuzzyTemplateLabelElement:
    """Generate a FuzzyTemplateLabelElement from a label and an element json dictionary.

    :param label: the label string for the label element
    :type label: str
    :param element_info: a dictionary containing the properties of the template label element
    :type element_info: dict
    :return: a fuzzy template label element
    :rtype: FuzzyTemplateLabelElement
    """
    required = element_info["required"] if "required" in element_info else False
    cardinality = element_info["cardinality"] if "cardinality" in element_info else "single"
    next_label = element_info["next_label"] if "next_label" in element_info else None
    next_distance_max = element_info["next_distance_max"] if "next_distance_max" in element_info else None
    variable = element_info["variable"] if "variable" in element_info else False
    return FuzzyTemplateLabelElement(label, required, cardinality, next_label, next_distance_max, variable)


def generate_group_from_json(element_info: dict,
                             group_elements: List[FuzzyTemplateElement]) -> FuzzyTemplateGroupElement:
    """Generate a FuzzyTemplateGroupElement from a element json dictionary and a list of group elements.

    :param element_info: a dictionary containing the properties of the template group element
    :type element_info: dict
    :param group_elements: a list of fuzzy template elements that are part of the group element
    :type group_elements: List[FuzzyTemplateElement]
    :return: a fuzzy template group element
    :rtype: FuzzyTemplateGroupElement
    """
    label = element_info["label"] if "label" in element_info else None
    ordered = element_info["ordered"] if "ordered" in element_info else True
    required = element_info["required"] if "required" in element_info else False
    return FuzzyTemplateGroupElement(group_elements, label=label, ordered=ordered, required=required)


class FuzzyTemplate:

    def __init__(self, phrase_model: PhraseModel,
                 template_json: Union[List[str], List[dict], Dict[str, Union[str, dict]]],
                 ignore_unknown: bool = False, ordered: bool = False):
        """A fuzzy search template to register phrases from a phrase model as elements of the template.
        The template can be used in combination with a fuzzy template searcher to find a given set
        of phrases (the template) within a certain range of text. The order, cardinality of elements
        and the distance between them can be specified. Elements can be grouped into ordered or unordered
        blocks of elements. Element groups can be hierarchical.

        :param phrase_model: a PhraseModel object with phrases that correspond to the element label
        in the template.
        :type phrase_model: PhraseModel
        :param template_json: a dictionary of template groups or elements to be registered as part of the template
        :type template_json: Union[List[str], List[dict], Dict[str, Union[str, dict]]]
        :param ignore_unknown: whether to ignore elements with labels that are not in the phrase model
        :type ignore_unknown: bool
        """
        self.phrase_model = phrase_model
        self.ordered = ordered
        self.labels: Set[str] = set()
        self.elements: Set[FuzzyTemplateLabelElement] = set()
        self.groups: Set[FuzzyTemplateGroupElement] = set()
        self.required: Set[FuzzyTemplateLabelElement] = set()
        self.label_element_index: Dict[str, FuzzyTemplateLabelElement] = {}
        self.group_element_index: Dict[str, FuzzyTemplateGroupElement] = {}
        self.root_element = None
        self.ignore_unknown = ignore_unknown
        self.register_template(template_json)

    def __repr__(self):
        return f"FuzzyTemplate(labels={self.labels}, required={self.get_required_labels()})"

    def get_label_phrases(self, label: str) -> List[Phrase]:
        """Return a list of phrases that have a given label.

        :param label: a phrase label for phrases in the registered phrase_model
        :type label: str
        :return: a list of phrases from the registered phrase model that have a given phrase
        :rtype: List[Phrase]
        """
        if label not in self.phrase_model.is_label_of:
            return []
        return [self.phrase_model.phrase_index[phrase_string] for phrase_string in self.phrase_model.is_label_of[label]]

    def has_label(self, label: Union[str, List[str]], ) -> bool:
        """Check if the template has label elements with a given label or list of label (any or all).

        :param label: a fuzzy element label
        :type label: Union[str, List[str]]
        :return: whether the label corresponds to any registered element(s)
        :rtype: bool
        """
        if isinstance(label, list):
            for label_item in label:
                if label_item in self.label_element_index:
                    return True
            return False
        else:
            return label in self.label_element_index

    def has_group(self, group: str) -> bool:
        """Check if the template has group elements with a given group name.

        :param group: a fuzzy element group
        :type group: str
        :return: whether the group corresponds to any registered element(s)
        :rtype: bool
        """
        return group in self.group_element_index

    def get_element(self, element_label: str) -> Union[None, FuzzyTemplateLabelElement, FuzzyTemplateGroupElement]:
        """Return the element corresponding to a given label.

        :param element_label: a fuzzy element label
        :type element_label: str
        :return: the element corresponding to the label or None if label is unknown
        :rtype: Union[FuzzyTemplateElement]
        """
        if self.has_label(element_label):
            return self.label_element_index[element_label]
        elif self.has_group(element_label):
            return self.group_element_index[element_label]
        else:
            return None

    def register_template(self, template_json: Union[List[str], List[dict], Dict[str, Union[str, dict]]]) -> None:
        """Register a list of elements as a fuzzy template. Each element contains a label
        that corresponds to at least one phrase in the registered phrase model.

        :param template_json: a dictionary of template groups or elements to be registered as part of the template
        :type template_json: Union[List[str], List[dict], Dict[str, Union[str, dict]]]
        """
        if isinstance(template_json, list):
            # if the template is a list, turn it into a dictionary first
            elements: List[Dict[str, any]] = []
            for ele in template_json:
                if isinstance(ele, str):
                    ele = {"label": ele, "type": "label"}
                elements.append(ele)
            template_json = {
                "type": "group",
                "elements": elements
            }
        if "type" not in template_json:
            template_json["type"] = "group" if "elements" in template_json else "label"
        if template_json["type"] == "label":
            # make sure the top level is always a group
            template_json = {"type": "group", "elements": [template_json]}
        elif template_json["type"] != "group":
            raise ValueError("element type must be 'label' or 'group'.")
        self.root_element = self.parse_group_element(template_json)

    def get_required_elements(self) -> List[FuzzyTemplateLabelElement]:
        """Return all required elements in the template.

        :return: the list of labels of required elements
        :rtype: List[FuzzyTemplateElement]
        """
        return [element for element in self.required]

    def get_required_labels(self) -> List[str]:
        """Return the labels of all required elements in the template.

        :return: the list of labels of required elements
        :rtype: List[str]
        """
        return [element.label for element in self.required]

    def get_elements_by_cardinality(self, cardinality: str = "single") -> List[FuzzyTemplateLabelElement]:
        """Return all template elements with a given cardinality.

        :param cardinality: a cardinality type ('single' or 'multi')
        :type cardinality: str
        :return: the list of labels of elements with a given cardinality
        :rtype: List[str]
        """
        return [element for element in self.elements if element.cardinality == cardinality]

    def get_labels_by_cardinality(self, cardinality: str = "single") -> List[str]:
        """Return the labels of all template elements with a given cardinality.

        :param cardinality: a cardinality type ('single' or 'multi')
        :type cardinality: str
        :return: the list of labels of elements with a given cardinality
        :rtype: List[str]
        """
        return [element.label for element in self.elements if element.cardinality == cardinality]

    def parse_label_element(self, label_info: Union[str, Dict[str, any]]) -> Union[FuzzyTemplateLabelElement, None]:
        """Parse a label element dictionary/JSON object into a fuzzy template label element.

        :param label_info: a dictionary containing the properties of the template label element
        :type label_info: dict
        :return: a fuzzy template label element, or None if the label is not used in the phrase model
        :rtype: FuzzyTemplateLabelElement
        """
        if isinstance(label_info, str):
            label_info = {"label": label_info}
        if "label" not in label_info:
            print("element json without label:", label_info)
        label = label_info["label"]
        if not self.phrase_model.is_label(label):
            if "variable" in label_info and label_info["variable"] is True:
                pass
            elif self.ignore_unknown:
                print("skipping unknown phrase label", label)
                return None
            else:
                raise ValueError(f"label '{label}' does not correspond to any phrase in registered phrase model")
        element = generate_label_from_json(label, label_info)
        # register the label element as part of the template
        self.label_element_index[element.label] = element
        self.labels.add(element.label)
        if element.required:
            # keep track of all label elements that are required
            self.required.add(element)
        return element

    def parse_group_element(self, group_info: Dict[str, any]) -> FuzzyTemplateGroupElement:
        """Parse a group element dictionary/JSON object into a fuzzy template group element.

        :param group_info: a dictionary containing the properties of the template group element
        :type group_info: dict
        :return: a fuzzy template group element
        :rtype: FuzzyTemplateGroupElement
        """
        group_elements = []
        for element_info in group_info["elements"]:
            if isinstance(element_info, str):
                # if element is a string, it must be a label element
                element_info = {"label": element_info, "type": "label"}
            elif "type" not in element_info:
                element_info["type"] = "group" if "elements" in element_info else "label"
            if element_info["type"] == "group":
                element = self.parse_group_element(element_info)
            else:
                element = self.parse_label_element(element_info)
                if element is None:
                    continue
            group_elements.append(element)
        group = generate_group_from_json(group_info, group_elements)
        # register the group element as part of the template
        self.groups.add(group)
        if group.label:
            self.group_element_index[group.label] = group
        return group



from unittest import TestCase
import string
from fuzzy_search.fuzzy_match import PhraseMatch, PhraseMatchInContext, adjust_match_offsets
from fuzzy_search.fuzzy_phrase import Phrase
from fuzzy_search.fuzzy_match import adjust_match_start_offset, adjust_match_end_offset
from fuzzy_search.fuzzy_match import map_string


class TestFuzzyMatch(TestCase):

    def test_map_string_maps_word_string(self):
        affix_string = "test"
        mapped_string = map_string(affix_string, string.punctuation)
        self.assertEqual(mapped_string, "wwww")

    def test_map_string_maps_space_string(self):
        affix_string = "     "
        mapped_string = map_string(affix_string, string.punctuation)
        self.assertEqual(mapped_string, "sssss")

    def test_map_string_maps_mixed_string(self):
        affix_string = "a test"
        mapped_string = map_string(affix_string, string.punctuation)
        self.assertEqual(mapped_string, "wswwww")

    def test_adjust_start(self):
        text = {"text": "this text contains some words"}
        match_string = "contains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, start)

    def test_adjust_start_returns_when_in_middle_of_word(self):
        text = {"text": "this text concontains some words"}
        match_string = "contains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, None)

    def test_adjust_start_shifts_to_previous_character(self):
        text = {"text": "this text contains some words"}
        match_string = "ontains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, start-1)

    def test_adjust_start_shifts_to_second_previous_characters(self):
        text = {"text": "this text contains some words"}
        match_string = "ntains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, start-2)

    def test_adjust_start_does_not_shift_to_third_previous_character(self):
        text = {"text": "this text contains some words"}
        match_string = "tains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, None)

    def test_adjust_start_shifts_to_next_character(self):
        text = {"text": "this text contains some words"}
        match_string = " contains"
        start = text["text"].index(match_string)
        adjusted_start = adjust_match_start_offset(text, match_string, start)
        self.assertEqual(adjusted_start, start+1)

    def test_adjust_end(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains"
        match_string = "contains"
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, end)

    def test_adjust_end_shifts_back_one_when_ending_with_whitespace(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains"
        match_string = "contains "
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, end-1)

    def test_adjust_end_shifts_back_two_when_ending_with_whitespace_and_char(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains"
        match_string = "contains s"
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, end-2)

    def test_adjust_end_shifts_back_one_when_phrase_ends_with_whitespace(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains "
        match_string = "contains s"
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, end-1)

    def test_adjust_end_does_not_shift_when_end_middle_of_next_word(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains"
        match_string = "contains so"
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, None)

    def test_adjust_end_shifts_to_end_of_next_word(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains som"
        match_string = "contains som"
        end = text["text"].index(match_string) + len(match_string)
        adjusted_end = adjust_match_end_offset(phrase_string, match_string, text, end, string.punctuation)
        self.assertEqual(adjusted_end, end+1)

    def test_adjust_boundaries_removes_surrounding_whitespace(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains"
        candidate_string = " contains "
        start = text["text"].index(candidate_string)
        end = text["text"].index(candidate_string) + len(candidate_string)
        adjusted_match = adjust_match_offsets(phrase_string, candidate_string, text, start, end)
        self.assertEqual(adjusted_match["match_string"], phrase_string)

    def test_adjust_boundaries_finds_word_boundary(self):
        text = {"text": "this text contains some words"}
        phrase_string = "contains som"
        candidate_string = " contains som"
        start = text["text"].index(candidate_string)
        end = text["text"].index(candidate_string) + len(candidate_string)
        adjusted_match = adjust_match_offsets(phrase_string, candidate_string, text, start, end)
        self.assertNotEqual(adjusted_match, None)
        self.assertEqual(adjusted_match["match_string"], phrase_string + "e")


class TestMatchInContext(TestCase):

    def setUp(self) -> None:
        self.text = "This string contains test text."
        self.phrase = Phrase("test")
        self.match = PhraseMatch(self.phrase, self.phrase, "test", 21)

    def test_make_match_in_context(self):
        match_in_context = PhraseMatchInContext(self.match, self.text)
        self.assertEqual(match_in_context.context_start, 1)

    def test_context_is_adjustable(self):
        match_in_context = PhraseMatchInContext(self.match, self.text, prefix_size=10)
        self.assertEqual(match_in_context.context_start, 11)

    def test_context_contains_text_from_doc(self):
        match_in_context = PhraseMatchInContext(self.match, self.text, prefix_size=30)
        self.assertEqual(match_in_context.context, self.text)


class TestMatchAnnotation(TestCase):

    def setUp(self) -> None:
        self.text = "This string contains test text."
        self.phrase = Phrase("test")
        self.match = PhraseMatch(self.phrase, self.phrase, match_string="test", match_offset=21,
                                 text_id='test_id', match_label='test_label')

    def test_match_as_web_anno_has_multiple_bodies(self):
        match_anno = self.match.as_web_anno()
        self.assertEqual(len(match_anno['body']), 3)

    def test_match_as_web_anno_skips_correction_with_exact_match(self):
        match_anno = self.match.as_web_anno()
        label_body = None
        for body in match_anno['body']:
            if body['purpose'] == 'correcting':
                label_body = body
        self.assertEqual(label_body, None)

    def test_match_as_web_anno_has_label(self):
        match_anno = self.match.as_web_anno()
        label_body = None
        for body in match_anno['body']:
            if body['purpose'] == 'classifying':
                label_body = body
        self.assertNotEqual(label_body, None)
        self.assertEqual(label_body['value'], self.match.label)

    def test_match_in_context_as_web_anno_has_multiple_selectors(self):
        match_in_context = PhraseMatchInContext(self.match, self.text, prefix_size=10)
        match_anno = match_in_context.as_web_anno()
        self.assertEqual(isinstance(match_anno['target']['selector'], list), True)
        self.assertEqual(len(match_anno['target']['selector']), 2)

    def test_match_in_context_as_web_anno_has_text_quote_selector(self):
        match_in_context = PhraseMatchInContext(self.match, self.text, prefix_size=10)
        match_anno = match_in_context.as_web_anno()
        quote_selector = None
        for selector in match_anno['target']['selector']:
            if selector['type'] == 'TextQuoteSelector':
                quote_selector = selector
        self.assertNotEqual(quote_selector, None)
        self.assertEqual(quote_selector['prefix'], self.text[self.match.offset-10:self.match.offset])



from typing import List, Union

from fuzzy_search.fuzzy_match import PhraseMatch, PhraseMatchInContext
from fuzzy_search.fuzzy_phrase_searcher import FuzzyPhraseSearcher


class FuzzyContextSearcher(FuzzyPhraseSearcher):

    def __init__(self, config: Union[dict, None] = None):
        super().__init__(config)
        self.context_size = 100
        if config is not None:
            self.configure_context(config)

    def configure_context(self, config: dict) -> None:
        """Configure the context searcher.

        :param config: a dictionary with configuration parameters to override the defaults
        :type config: dict
        """
        super().configure(config)
        if "context_size" in config:
            self.context_size = config["context_size"]

    def add_match_context(self, match: PhraseMatch, text: Union[str, dict], context_size: Union[None, int] = None,
                          prefix_size: Union[None, int] = None,
                          suffix_size: Union[None, int] = None) -> PhraseMatchInContext:
        """Add context to a given match and its corresponding text document.

        :param match: a phrase match object
        :type match: PhraseMatch
        :param text: the text that the match was taken from
        :type text: Union[str, dict]
        :param context_size: the size of the pre- and suffix window
        :type context_size: int
        :param prefix_size: size of the prefix context
        :type prefix_size: Union[None, int]
        :param suffix_size: size of the suffix context
        :type suffix_size: Union[None, int]
        :return: the phrase match object with context
        :rtype: PhraseMatchInContext
        """
        if context_size is None:
            context_size = self.context_size
        prefix_size = prefix_size if prefix_size is not None else context_size
        suffix_size = suffix_size if suffix_size is not None else context_size
        return PhraseMatchInContext(match, text, prefix_size=prefix_size, suffix_size=suffix_size)

    def find_matches(self, text: Union[str, dict],
                     use_word_boundaries: Union[None, bool] = None,
                     allow_overlapping_matches: bool = True,
                     include_variants: bool = None,
                     filter_distractors: bool = None,
                     prefix_size: Union[None, int] = None,
                     suffix_size: Union[None, int] = None,
                     skip_exact_matching: bool = None) -> List[PhraseMatchInContext]:
        """Find fuzzy matches for registered phrases and add context around match string. This extends
        the find_matches function of the FuzzyPhraseSearcher by adding local context to each match.

        :param text: the text (string or dictionary with 'text' property) to find fuzzy matching phrases in.
        :type text: Union[str, Dict[str, str]]
        :param use_word_boundaries: use word boundaries in determining match boundaries
        :type use_word_boundaries: bool
        :param allow_overlapping_matches: boolean flag for whether to allow matches to overlap in their text ranges
        :type allow_overlapping_matches: bool
        :param include_variants: boolean flag for whether to include phrase variants for finding matches
        :type include_variants: bool
        :param filter_distractors: boolean flag for whether to remove phrase matches that better match distractors
        :type filter_distractors: bool
        :param prefix_size: the size of the prefix context window
        :type prefix_size: Union[None, int]
        :param suffix_size: the size of the suffix context window
        :type suffix_size: Union[None, int]
        :param skip_exact_matching: boolean flag whether to skip the exact matching step
        :type skip_exact_matching: Union[None, bool]
        :return: a list of phrases matches with text surrounding the match string
        :rtype: PhraseMatchInContext
        """
        matches = super().find_matches(text, use_word_boundaries=use_word_boundaries,
                                       allow_overlapping_matches=allow_overlapping_matches,
                                       include_variants=include_variants, filter_distractors=filter_distractors,
                                       skip_exact_matching=skip_exact_matching)
        return [self.add_match_context(match, text, prefix_size=prefix_size,
                                       suffix_size=suffix_size) for match in matches]

    def find_matches_in_context(self, match_in_context: PhraseMatchInContext,
                                use_word_boundaries: Union[None, bool] = None,
                                include_variants: Union[None, bool] = None,
                                filter_distractors: Union[None, bool] = None) -> List[PhraseMatch]:
        """Use a MatchInContext object to find other phrases in the context of that match.

        :param match_in_context: a match phrase with context from the text that the match was taken from
        :type match_in_context: PhraseMatchInContext
        :param use_word_boundaries: boolean whether to adjust match strings to word boundaries
        :type use_word_boundaries: bool
        :param include_variants: boolean whether to include variants of phrases in matching
        :type include_variants: bool
        :param filter_distractors: boolean whether to remove matches that are closer to distractors
        :type filter_distractors: bool
        :return: a list of match objects
        :rtype: List[PhraseMatch]
        """
        # override searcher configuration if a boolean is passed
        use_word_boundaries = self.use_word_boundaries if use_word_boundaries is None else use_word_boundaries
        include_variants = self.include_variants if include_variants is None else include_variants
        filter_distractors = self.filter_distractors if filter_distractors is None else filter_distractors
        # search the context of the original match
        context_matches = self.find_matches(match_in_context.context, use_word_boundaries=use_word_boundaries,
                                            include_variants=include_variants,
                                            filter_distractors=filter_distractors)
        # recalculate the match offset with respect to the original text
        for match in context_matches:
            match.offset += match_in_context.context_start
            match.end += match_in_context.context_start
        return context_matches


if __name__ == "__main__":
    sample_text = "A'nthony van der Truyn en Adriaen Bosman, Makelaers tot Rotterdam, prefenteren," + \
                  "uyt de Hint te verkopen etfn curieufc Party opreckw ?al somfl'e Schalyen of Leyen."



from unittest import TestCase
from fuzzy_search.fuzzy_string import make_ngrams, score_char_overlap, score_char_overlap_ratio
from fuzzy_search.fuzzy_string import score_ngram_overlap, score_ngram_overlap_ratio
from fuzzy_search.fuzzy_string import score_levenshtein_distance, score_levenshtein_similarity_ratio


class Test(TestCase):

    ###############
    # make_ngrams #
    ###############

    def test_make_ngrams_rejects_non_string_text(self):
        error = None
        try:
            make_ngrams(1, 1.5)
        except TypeError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_make_ngrams_rejects_non_integer_size(self):
        error = None
        try:
            make_ngrams('test', 1.5)
        except TypeError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_make_ngrams_rejects_negative_size(self):
        error = None
        try:
            make_ngrams('test', -1)
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_make_ngrams_accepts_positive_integer(self):
        error = None
        try:
            make_ngrams('test', 1)
        except ValueError as err:
            error = err
        self.assertEqual(error, None)

    def test_make_ngrams_handles_size_one_correctly(self):
        text = 'test'
        ngram_size = 1
        ngrams = make_ngrams(text, ngram_size)
        num_ngrams = len(text) + 3 - ngram_size
        self.assertEqual(len(ngrams), num_ngrams)

    def test_make_ngrams_handles_size_two_correctly(self):
        text = 'test'
        ngram_size = 2
        ngrams = make_ngrams(text, ngram_size)
        num_ngrams = len(text) + 3 - ngram_size
        self.assertEqual(len(ngrams), num_ngrams)

    def test_make_ngrams_rejects_integer_larger_than_text_length(self):
        ngrams = make_ngrams('test', 10)
        self.assertEqual(len(ngrams), 0)

    #######################
    # score_ngram_overlap #
    #######################

    def test_score_ngram_overlap_is_num_ngrams_for_self_comparison(self):
        text = 'test'
        ngram_size = 2
        ngrams = make_ngrams(text, ngram_size)
        overlap = score_ngram_overlap(text, text, ngram_size)
        self.assertEqual(overlap, len(ngrams))

    def test_score_ngram_overlap_is_zero_for_comparison_with_empty(self):
        text = 'test'
        ngram_size = 2
        overlap = score_ngram_overlap(text, '', ngram_size)
        self.assertEqual(overlap, 0)

    ######################
    # score_char_overlap #
    ######################

    def test_score_char_overlap_with_self_is_len_of_self(self):
        text = 'test'
        overlap = score_char_overlap(text, text)
        self.assertEqual(overlap, len(text))

    def test_score_char_overlap_with_smaller_word_is_smaller_than_len_of_self(self):
        text = 'ttttt'
        overlap = score_char_overlap(text, text[:4])
        self.assertNotEqual(overlap, len(text))

    ################################
    # score_levenshtein_similarity #
    ################################

    def test_score_levenshtein_similarity_with_self_is_len_of_self(self):
        text = 'test'
        similarity = score_levenshtein_similarity_ratio(text, text)
        self.assertEqual(similarity, 1)



from unittest import TestCase
from typing import Generator
from fuzzy_search.fuzzy_phrase import text2skipgrams, Phrase


class Test(TestCase):

    # text2skipgrams

    def test_text2skipgrams_rejects_negative_ngram_size(self):
        error = None
        try:
            next(text2skipgrams("test", ngram_size=-1, skip_size=2))
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_text2skipgrams_rejects_negative_skip_size(self):
        error = None
        try:
            next(text2skipgrams("test", ngram_size=2, skip_size=-1))
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_text2skipgrams_accepts_positive_ngram_size(self):
        skip_iter = text2skipgrams("test", ngram_size=2, skip_size=2)
        self.assertEqual(isinstance(skip_iter, Generator), True)

    def test_skipgrams_have_correct_length(self):
        skipgrams = [skipgram for skipgram in text2skipgrams("test", ngram_size=2, skip_size=2)]
        self.assertEqual(skipgrams[2].length, 4)


class TestFuzzyPhrase(TestCase):

    def test_fuzzy_phrase_accepts_phrase_as_string(self):
        phrase = Phrase("some phrase")
        self.assertEqual(phrase.phrase_string, "some phrase")

    def test_fuzzy_phrase_accepts_phrase_as_dict(self):
        phrase = Phrase({"phrase": "some phrase"})
        self.assertEqual(phrase.phrase_string, "some phrase")

    def test_fuzzy_phrase_can_set_max_offset(self):
        phrase = Phrase({"phrase": "some phrase", "max_offset": 3})
        self.assertEqual(phrase.max_offset, 3)

    def test_fuzzy_phrase_can_set_max_end(self):
        phrase = Phrase({"phrase": "some phrase", "max_offset": 3})
        self.assertEqual(phrase.max_end, 3 + len("some_phrase"))

    def test_fuzzy_phrase_cannot_set_negative_max_offset(self):
        error = None
        try:
            Phrase({"phrase": "some phrase", "max_offset": -3})
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)

    def test_fuzzy_phrase_accepts_phrase_with_valid_string_label(self):
        phrase = Phrase({"phrase": "some phrase", "label": "some_label"})
        self.assertEqual(phrase.label, "some_label")

    def test_fuzzy_phrase_accepts_phrase_with_valid_list_of_strings_label(self):
        phrase = Phrase({"phrase": "some phrase", "label": ["some_label", "other_label"]})
        self.assertEqual(isinstance(phrase.label, list), True)
        self.assertEqual("other_label" in phrase.label, True)

    def test_fuzzy_phrase_rejects_phrase_with_invalid_label(self):
        error = None
        try:
            Phrase({"phrase": "some phrase", "label": {"complex": "label"}})
        except ValueError as err:
            error = err
        self.assertNotEqual(error, None)



from unittest import TestCase
from fuzzy_search.fuzzy_phrase_model import PhraseModel


class Test(TestCase):

    def test_making_empty_phrase_model(self):
        phrase_model = PhraseModel()
        self.assertNotEqual(phrase_model, None)

    def test_making_phrase_model_with_list_of_keyword_strings(self):
        phrases = ["test"]
        phrase_model = PhraseModel(phrases=phrases)
        self.assertEqual(phrase_model.has_phrase(phrases[0]), True)

    def test_making_phrase_model_with_list_of_phrase_dictionaries(self):
        phrases = [{"phrase": "test"}]
        phrase_model = PhraseModel(phrases=phrases)
        self.assertEqual(phrase_model.has_phrase(phrases[0]["phrase"]), True)

    def test_phrase_model_can_add_phrase(self):
        phrases = [{"phrase": "test"}]
        phrase_model = PhraseModel()
        phrase_model.add_phrases(["test"])
        self.assertEqual(phrase_model.has_phrase(phrases[0]["phrase"]), True)

    def test_phrase_model_can_remove_phrase(self):
        phrases = [{"phrase": "test"}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.remove_phrases(["test"])
        self.assertEqual(phrase_model.has_phrase(phrases[0]["phrase"]), False)

    def test_phrase_model_indexes_phrase_words(self):
        phrases = [{"phrase": "this is a test"}]
        phrase_model = PhraseModel(phrases=phrases)
        self.assertEqual(phrases[0]["phrase"] in phrase_model.word_in_phrase["test"], True)

    def test_can_add_label_to_phrase(self):
        phrases = [{"phrase": "test", "label": "some_value"}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_custom(phrases)
        self.assertEqual(phrase_model.has_label(phrases[0]["phrase"]), True)

    def test_can_add_label_as_list_to_phrase(self):
        phrases = [{"phrase": "test", "label": ["some_value", "other_value"]}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_custom(phrases)
        label = phrase_model.get_labels(phrases[0]["phrase"])
        self.assertEqual(isinstance(label, set), True)
        self.assertEqual(len(label), 2)

    def test_can_add_custom_key_value_pairs_to_phrase(self):
        phrases = [{"phrase": "test", "some_key": "some_value"}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_custom(phrases)
        self.assertEqual(phrase_model.has_custom(phrases[0]["phrase"], "some_key"), True)

    def test_can_add_variant_phrase(self):
        phrases = [{"phrase": "okay", "variants": ["OK"]}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_variants(phrases)
        self.assertEqual(phrase_model.variant_of("OK").phrase_string, phrases[0]["phrase"])

    def test_can_add_distractors(self):
        phrases = [{"phrase": "okay", "distractors": ["OK"]}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_variants(phrases)
        self.assertEqual(phrases[0]["phrase"] in phrase_model.is_distractor_of["OK"], True)

    def test_can_configure_ngram_size(self):
        phrases = [{"phrase": "test", "label": "some_value"}]
        phrase_model = PhraseModel(phrases=phrases, config={"ngram_size": 3})
        self.assertEqual(phrase_model.ngram_size, 3)
        self.assertEqual(phrase_model.phrase_index["test"].ngram_size, 3)

    def test_can_get_json_representation(self):
        phrases = [{"phrase": "okay", "variants": ["OK"]}]
        phrase_model = PhraseModel(phrases=phrases)
        phrase_model.add_variants(phrases)
        phrase_json = phrase_model.json
        self.assertEqual(phrase_json[0]['phrase'], phrases[0]['phrase'])
        self.assertEqual(phrase_json[0]['variants'][0], phrases[0]['variants'][0])




from typing import Dict, Generator, List, Set, Union
from collections import defaultdict
import json
import copy
import re

from fuzzy_search.fuzzy_phrase import Phrase


def as_phrase_object(phrase: Union[str, dict, Phrase], ngram_size: int = 2, skip_size: int = 2) -> Phrase:
    if isinstance(phrase, Phrase):
        return phrase
    if isinstance(phrase, dict):
        if not is_phrase_dict(phrase):
            print(phrase)
            raise KeyError("invalid phrase dictionary")
        return Phrase(phrase, ngram_size=ngram_size, skip_size=skip_size)
    if isinstance(phrase, str):
        return Phrase(phrase, ngram_size=ngram_size, skip_size=skip_size)
    else:
        raise TypeError('phrase must be of type string')


def is_phrase_dict(phrase_dict: Dict[str, Union[str, List[str]]]) -> bool:
    if not isinstance(phrase_dict, dict):
        return False
    if "phrase" not in phrase_dict:
        return False
    if not isinstance(phrase_dict["phrase"], str):
        return False
    if "variants" in phrase_dict:
        for variant in phrase_dict["variants"]:
            if not isinstance(variant, str):
                return False
    if "distractors" in phrase_dict:
        for distractor in phrase_dict["distractors"]:
            if not isinstance(distractor, str):
                return False
    if "labels" in phrase_dict:
        if isinstance(phrase_dict["labels"], str):
            return True
        if not isinstance(phrase_dict["labels"], list):
            return False
        for label in phrase_dict["labels"]:
            if not isinstance(label, str):
                return False
    return True


class PhraseModel:

    def __init__(self, phrases: Union[None, List[Union[str, Dict[str, Union[str, list]], Phrase]]] = None,
                 variants: Union[None, List[Union[Dict[str, List[str]], Phrase]]] = None,
                 phrase_labels: Union[None, List[Dict[str, str]]] = None,
                 distractors: Union[None, List[Union[Dict[str, List[str]], Phrase]]] = None,
                 model: Union[None, List[Dict[str, Union[str, list]]]] = None,
                 custom: Union[None, List[Dict[str, Union[str, int, float, list]]]] = None,
                 config: dict = None):
        if config is None:
            config = {}
        self.ngram_size = config["ngram_size"] if "ngram_size" in config else 2
        self.skip_size = config["skip_size"] if "skip_size" in config else 2
        self.phrase_index: Dict[str, Phrase] = {}
        # only register variants of known phrases
        self.variant_index: Dict[str, Phrase] = {}
        self.has_variants: Dict[str, Set[str]] = defaultdict(set)
        self.is_variant_of: Dict[str, str] = {}
        self.distractor_index: Dict[str, Phrase] = {}
        self.has_distractors: Dict[str, Set[str]] = defaultdict(set)
        self.is_distractor_of: Dict[str, Set[str]] = defaultdict(set)
        self.phrase_length_index: Dict[int, set] = defaultdict(set)
        self.variant_length_index: Dict[int, set] = defaultdict(set)
        self.has_labels: Dict[str, Set[str]] = defaultdict(set)
        self.is_label_of: Dict[str, Set[str]] = defaultdict(set)
        self.custom = {}
        self.word_in_phrase: Dict[str, Set[str]] = defaultdict(set)
        self.first_word_in_phrase: Dict[str, Dict[str, int]] = defaultdict(dict)
        self.phrase_type: Dict[str, Set[str]] = defaultdict(set)
        self.phrase_string_map: Dict[str, Phrase] = {}
        if phrases:
            self.add_phrases(phrases)
        if variants:
            self.add_variants(variants)
        if distractors:
            self.add_distractors(distractors)
        if phrase_labels:
            self.add_labels(phrase_labels)
        if model:
            self.add_model(model)
        if custom:
            self.add_custom(custom)

    def __repr__(self):
        """A phrase model to support fuzzy searching in OCR/HTR output."""
        return f"PhraseModel({json.dumps(self.json(), indent=2)})"

    def __str__(self):
        return self.__repr__()

    def add_model(self, model: List[Union[str, Dict[str, Union[str, list]]]]) -> None:
        """Add an entire model with list of phrase dictionaries.

        :param model: a list of phrase dictionaries
        :type model: List[Union[str, Dict[str, Union[str list]]]]
        :return: None
        :rtype: None
        """
        self.add_phrases(model)
        self.add_variants(model)
        self.add_distractors(model)
        self.add_labels(model)
        self.add_custom(model)

    @property
    def json(self) -> List[Dict[str, Union[str, List[str]]]]:
        """Return a JSON representation of the phrase model.

        :return: a JSON respresentation of the phrase model
        :rtype: List[Dict[str, Union[str, List[str]]]]
        """
        model_json: List[Dict[str, Union[str, List[str]]]] = []
        for phrase in self.phrase_index:
            entry = {'phrase': phrase}
            if phrase in self.has_variants:
                entry['variants'] = list(self.has_variants[phrase])
            if phrase in self.has_labels:
                entry['label'] = list(self.has_labels[phrase])
            if phrase in self.custom:
                entry['custom'] = self.custom[phrase]
            model_json += [entry]
        return model_json

    def add_phrase(self, phrase: Phrase) -> None:
        """Add a phrase to the model as main phrase.

        :param phrase: a phrase to be added
        :type phrase: Phrase
        """
        self.phrase_string_map[phrase.phrase_string] = phrase
        self.phrase_type[phrase.phrase_string].add("phrase")
        self.phrase_index[phrase.phrase_string] = phrase
        self.phrase_length_index[len(phrase.phrase_string)].add(phrase.phrase_string)
        self.index_phrase_words(phrase)

    def add_variant(self, variant_phrase: Phrase, main_phrase: Phrase):
        """Add a phrase to the model as variant of a given main phrase.

        :param variant_phrase: a variant phrase to be added as variant of main_phrase
        :type variant_phrase: Phrase
        :param main_phrase: a main phrase that the variant phrase is a variant of
        :type main_phrase: Phrase
        """
        if variant_phrase.phrase_string not in self.phrase_string_map:
            self.phrase_string_map[variant_phrase.phrase_string] = variant_phrase
        self.variant_index[variant_phrase.phrase_string] = variant_phrase
        self.is_variant_of[variant_phrase.phrase_string] = main_phrase.phrase_string
        self.has_variants[main_phrase.phrase_string].add(variant_phrase.phrase_string)
        self.phrase_type[variant_phrase.phrase_string].add("variant")
        self.variant_length_index[len(variant_phrase.phrase_string)].add(variant_phrase.phrase_string)
        self.index_phrase_words(variant_phrase)

    def add_distractor(self, distractor_phrase: Phrase, main_phrase: Phrase):
        """Add a phrase to the model as distractor of a given main phrase.

        :param distractor_phrase: a distractor phrase to be added as distractor of main_phrase
        :type distractor_phrase: Phrase
        :param main_phrase: a main phrase that the distractor phrase is a distractor of
        :type main_phrase: Phrase
        """
        if distractor_phrase.phrase_string not in self.phrase_string_map:
            self.phrase_string_map[distractor_phrase.phrase_string] = distractor_phrase
        self.distractor_index[distractor_phrase.phrase_string] = distractor_phrase
        self.is_distractor_of[distractor_phrase.phrase_string].add(main_phrase.phrase_string)
        self.has_distractors[main_phrase.phrase_string].add(distractor_phrase.phrase_string)
        self.phrase_type[distractor_phrase.phrase_string].add("distractor")
        self.index_phrase_words(distractor_phrase)

    def remove_phrase(self, phrase: Phrase):
        """Remove a main phrase from the model, including its connections to any variant and distractor phrases.

        :param phrase: a phrase that is registered as a main phrase
        :type phrase: Phrase
        """
        # first check if phrase is registered in this phrase model
        if phrase.phrase_string not in self.phrase_index:
            raise ValueError(f"{phrase.phrase_string} is not registered as a main phrase")
        # remove phrase from the type index
        self.phrase_type[phrase.phrase_string].remove("phrase")
        # remove the phrase string from the main phrase index
        del self.phrase_index[phrase.phrase_string]
        # remove the phrase from the phrase length index
        self.phrase_length_index[len(phrase.phrase_string)].remove(phrase.phrase_string)
        if len(self.phrase_type[phrase.phrase_string]) == 0:
            # if the phrase string is not registered as another type (variant or distractor)
            # remove the phrase words from the word_to_phrase index
            self.remove_phrase_words(phrase)
        # if the phrase has variants, remove those as well
        if phrase.phrase_string in self.has_variants:
            for variant_string in self.has_variants:
                variant_phrase = self.variant_index[variant_string]
                self.remove_variant(variant_phrase)
        # if the phrase has distractors, remove its connections with them as well
        if phrase.phrase_string in self.has_distractors:
            for distractor_string in self.has_distractors:
                distractor_phrase = self.distractor_index[distractor_string]
                if len(self.is_distractor_of[distractor_string]) > 1:
                    self.is_distractor_of[distractor_string].remove(phrase.phrase_string)
                else:
                    # if the distractor is only connected to this phrase, remove the distractor as well
                    self.remove_distractor(distractor_phrase)
            del self.has_distractors[phrase.phrase_string]

    def remove_variant(self, variant_phrase: Phrase) -> None:
        """Remove a variant phrase from the model, including its connection to the phrase it is a
        variant of.

        :param variant_phrase: a phrase that is registered as a variant of one or more main phrases
        :type variant_phrase: Phrase
        """
        # first check if variant phrase is registered as a variant
        if variant_phrase.phrase_string not in self.is_variant_of:
            raise ValueError(f"{variant_phrase.phrase_string} is not registered as a variant")
        # remove variant from the type index
        self.phrase_type[variant_phrase.phrase_string].remove("variant")
        # if that is the only type of the phrase, remove it from the word_to_phrase index
        if len(self.phrase_type[variant_phrase.phrase_string]) == 0:
            self.remove_phrase_words(variant_phrase)
        # remove the variant from the variant index
        del self.variant_index[variant_phrase.phrase_string]
        # remove the variant from the phrase length index
        self.variant_length_index[len(variant_phrase.phrase_string)].remove(variant_phrase.phrase_string)
        # remove its connection with its main phrase
        main_phrase_string = self.is_variant_of[variant_phrase.phrase_string]
        del self.is_variant_of[variant_phrase.phrase_string]
        self.has_variants[main_phrase_string].remove(variant_phrase.phrase_string)
        # if this was the only variant of the main phrase, remove the main phrase from the has_variants index
        if len(self.has_variants[main_phrase_string]) == 0:
            del self.has_variants[main_phrase_string]

    def remove_distractor(self, distractor_phrase: Phrase) -> None:
        """Remove a distractor phrase from the model, including its connection to the phrase it is a
        distractor of.

        :param distractor_phrase: a phrase that is registered as a distractor of one or more main phrases
        :type distractor_phrase: Phrase
        """
        if distractor_phrase.phrase_string not in self.is_distractor_of:
            raise ValueError(f"{distractor_phrase.phrase_string} is not registered as a distractor")
        self.phrase_type[distractor_phrase.phrase_string].remove("distractor")
        if len(self.phrase_type[distractor_phrase.phrase_string]) == 0:
            self.remove_phrase_words(distractor_phrase)
        del self.distractor_index[distractor_phrase.phrase_string]
        for main_phrase_string in self.is_distractor_of[distractor_phrase.phrase_string]:
            self.has_distractors[main_phrase_string].remove(distractor_phrase.phrase_string)
            if len(self.has_distractors[main_phrase_string]) == 0:
                del self.has_distractors[main_phrase_string]
        del self.is_distractor_of[distractor_phrase.phrase_string]

    def add_phrases(self, phrases: List[Union[str, Dict[str, Union[str, List[str]]], Phrase]]) -> None:
        """Add a list of phrases to the phrase model. Phrases must be either:
        - a list of strings
        - a list of dictionaries with property 'phrase' and the phrase as a string value
        - a list of Phrase objects

        :param phrases: a list of phrases
        :type phrases: List[Union[str, Dict[str, Union[str, List[str]]]]]
        """
        for phrase in phrases:
            phrase = as_phrase_object(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            self.add_phrase(phrase)
        # if phrases is a dictionary with possible variants, distractors, labels and custom metadata
        # per phrase, add those variants and distractors
        phrase_dicts = [phrase for phrase in phrases if isinstance(phrase, dict)]
        self.add_variants(phrase_dicts)
        self.add_distractors(phrase_dicts)
        self.add_custom(phrase_dicts)
        self.add_labels(phrase_dicts)

    def remove_phrases(self, phrases: List[Union[str, Dict[str, Union[str, List[str]]], Phrase]]):
        """Remove a list of phrases from the phrase model. If it has any registered spelling variants,
        remove those as well.

        :param phrases: a list of phrases/keyphrases
        :type phrases: List[Union[str, Dict[str, Union[str, List[str]]]]]
        """
        for phrase in phrases:
            phrase = as_phrase_object(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if phrase.phrase_string not in self.phrase_index:
                raise KeyError(f"Unknown phrase: {phrase.phrase_string}")
            self.remove_phrase(phrase)

    def get_phrases_by_max_length(self, max_length: int,
                                  include_variants: bool = False) -> Generator[Phrase, None, None]:
        """Return all phrase in the phrase model that are no longer than a given length.

        :param max_length: the maximum length of phrases to be returned
        :type max_length: int
        :param include_variants: whether to include variants
        :return: a generator that yield phrases
        :rtype: Generator[Phrase, None, None]
        """
        for phrase_length in self.phrase_length_index:
            if phrase_length > max_length:
                break
            for phrase_string in self.phrase_length_index[phrase_length]:
                yield self.phrase_index[phrase_string]
            if include_variants:
                for phrase_string in self.variant_length_index[phrase_length]:
                    yield self.variant_index[phrase_string]

    def get_phrases(self) -> List[Phrase]:
        """Return a list of all registered phrases.

        :return: a list of all registered phrases
        :rtype: List[Phrase]
        """
        return list(self.phrase_index.values())

    def has_phrase(self, phrase: Union[str, Dict[str, any], Phrase]) -> bool:
        """Check if phrase is registered in phrase_model.

        :param phrase: a phrase string
        :type phrase: Union[str, Dict[str, any], Phrase]
        :return: a boolean indicating whether phrase is registered
        :rtype: bool
        """
        phrase = as_phrase_object(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
        return phrase.phrase_string in self.phrase_index

    def add_variants(self, variants: List[Dict[str, Union[str, List[str]]]],
                     add_new_phrases: bool = True) -> None:
        """Add variants of a phrase. If the phrase is not registered, add it to the set.
        - input is a list of dictionaries:
        variants = [{'phrase': 'some phrase', 'variants': ['some variant', 'some other variant']}]

        :param variants: a list of phrase dictionaries with 'variant' property
        :type variants: List[Dict[str, Union[str, List[str]]]]
        :param add_new_phrases: a Boolean to indicate if unknown phrases should be added
        :type add_new_phrases: bool
        """
        for phrase_dict in variants:
            main_phrase = as_phrase_object(phrase_dict, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if main_phrase.phrase_string not in self.phrase_index:
                if add_new_phrases:
                    self.add_phrase(main_phrase)
                else:
                    continue
            if "variants" not in main_phrase.metadata:
                continue
            for variant_phrase_string in main_phrase.metadata["variants"]:
                variant_phrase = as_phrase_object(variant_phrase_string, ngram_size=self.ngram_size,
                                                  skip_size=self.skip_size)
                self.add_variant(variant_phrase, main_phrase)

    def remove_variants(self, variants: Union[List[Union[str, Phrase]], None] = None,
                        variants_of_phrase: Union[str, Phrase, None] = None):
        """Remove a list of spelling variants of a phrase.

        :param variants: a list of variant strings or variant phrase objects to remove
        :type variants: Union[List[str, Phrase]], None]
        :param variants_of_phrase: an optional phrase string or phrase object for which all variants are to be removed
        :type variants_of_phrase: Union[str, Phrase, None]
        """
        if variants:
            for variant in variants:
                variant = as_phrase_object(variant, ngram_size=self.ngram_size, skip_size=self.skip_size)
                self.remove_variant(variant)
        if variants_of_phrase:
            main_phrase = as_phrase_object(variants_of_phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if main_phrase not in self.phrase_index:
                raise IndexError(f"{main_phrase.phrase_string} is not registered in this phrase model")
            if main_phrase not in self.has_variants:
                return None
            for variant_string in self.has_variants[main_phrase.phrase_string]:
                variant = as_phrase_object(variant_string, ngram_size=self.ngram_size, skip_size=self.skip_size)
                self.remove_variant(variant)

    def get_variants(self, phrases: List[str] = None) -> List[Dict[str, Union[str, List[str]]]]:
        """Return registered variants of a specific list of phrases or
        of all registered phrases (when no list of phrases is given).

        :param phrases: a list of registered phrase strings
        :type phrases: List[str]
        :return: a list of dictionaries of phrases and their variants
        :rtype: List[Dict[str, Union[str, List[str]]]]
        """
        if phrases is None:
            phrases = self.phrase_index.keys()
        phrases = [as_phrase_object(phrase, ngram_size=self.ngram_size, skip_size=self.skip_size) for phrase in phrases]
        return [{'phrase': phrase.phrase_string, 'variants': self.has_variants[phrase.phrase_string]}
                for phrase in phrases]

    def variant_of(self, variant: Union[str, Phrase]) -> Union[None, Phrase]:
        variant_string = variant.phrase_string if isinstance(variant, Phrase) else variant
        if variant_string in self.is_variant_of:
            phrase_string = self.is_variant_of[variant_string]
            return self.phrase_index[phrase_string]
        else:
            return None

    def variants(self, phrase: Union[str, Phrase]) -> Union[None, List[Phrase]]:
        """Return all variants of a given phrase.

        :param phrase: a phrase string or phrase object
        :type phrase: Union[str, Phrase]
        :return: a list of variants of the phrase or None if it doesn't have any
        :type: Union[None, List[Phrase]]
        """
        phrase_string = phrase.phrase_string if isinstance(phrase, Phrase) else phrase
        if phrase_string not in self.has_variants:
            return None
        else:
            return [self.variant_index[variant_string] for variant_string in self.has_variants[phrase_string]]

    def add_distractors(self, distractors: List[Dict[str, Union[str, List[str]]]],
                        add_new_phrases: bool = True) -> None:
        """Add distractors of a phrase. If the phrase is not registered, add it to the set.
        - input is a list of dictionaries:
        distractors = [{'phrase': 'some phrase', 'distractors': ['some distractor', 'some other distractor']}]

        :param distractors: a list of phrase dictionaries with 'distractor' property
        :type distractors: List[Dict[str, Union[str, List[str]]]]
        :param add_new_phrases: a Boolean to indicate if unknown phrases should be added
        :type add_new_phrases: bool
        """
        for phrase_dict in distractors:
            main_phrase = as_phrase_object(phrase_dict, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if main_phrase.phrase_string not in self.phrase_index:
                if add_new_phrases:
                    self.add_phrase(main_phrase)
                else:
                    continue
            if "distractors" not in main_phrase.metadata:
                continue
            for distractor_string in main_phrase.metadata["distractors"]:
                distractor_phrase = as_phrase_object(distractor_string, ngram_size=self.ngram_size,
                                                     skip_size=self.skip_size)
                self.add_distractor(distractor_phrase, main_phrase)

    def remove_distractors(self, distractors: Union[List[Union[str, Phrase]], None] = None,
                           distractors_of_phrase: Union[str, None] = None):
        """Remove a list of distractors of a phrase.
        - distractors: a list of dictionaries with phrases as key and the list of distractors to be removed as values
        distractors = [{'phrase': 'some phrase', 'distractors': ['distractor to remove', 'some other distractor']}]
        - phrase: remove all distractors of a given phrase

        :param distractors: an optional list of phrase dictionaries with 'distractors' property
        :type distractors: Union[List[Union[str, Phrase]], None]
        :param distractors_of_phrase: an optional string of a registered phrase for which all distractors are removed
        :type distractors_of_phrase: Union[str, None]
        """
        if distractors:
            for distractor in distractors:
                distractor = as_phrase_object(distractor, ngram_size=self.ngram_size, skip_size=self.skip_size)
                self.remove_distractor(distractor)
        if distractors_of_phrase:
            main_phrase = as_phrase_object(distractors_of_phrase, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if main_phrase not in self.phrase_index:
                raise IndexError(f"{main_phrase.phrase_string} is not registered in this phrase model")
            if main_phrase not in self.has_distractors:
                return None
            for distractor_string in self.has_distractors[main_phrase.phrase_string]:
                distractor = as_phrase_object(distractor_string, ngram_size=self.ngram_size, skip_size=self.skip_size)
                self.remove_distractor(distractor)

    def add_labels(self, phrase_labels: List[Dict[str, Union[str, list]]]):
        """Add a label to a phrase. This can be used to group phrases under the same label.
        - input is a list of phrase/label pair dictionaries:
        labels = [{'phrase': 'some phrase', 'label': 'some label'}]
        """
        for phrase_dict in phrase_labels:
            phrase = as_phrase_object(phrase_dict, ngram_size=self.ngram_size, skip_size=self.skip_size)
            if phrase.label is None:
                continue
            if phrase.phrase_string not in self.phrase_index:
                print(f'skipping label for unknown phrase {phrase}')
            for label in phrase.label_set:
                self.has_labels[phrase.phrase_string].add(label)
                self.is_label_of[label].add(phrase.phrase_string)

    def remove_labels(self, phrases: Union[List[Phrase], List[str]]) -> None:
        """Remove labels for known phrases.

        :param phrases: is a list of known phrases (either as Phrase objects or strings)
        :type phrases: Union[List[Phrase], List[str]]
        """
        for phrase in phrases:
            phrase_string = phrase if isinstance(phrase, str) else phrase.phrase_string
            if phrase_string not in self.phrase_index:
                raise TypeError(f'unknown phrase {phrase_string}')
            else:
                for label in self.has_labels[phrase_string]:
                    self.is_label_of[label].remove(phrase_string)
                    if len(self.is_label_of[label]) == 0:
                        del self.is_label_of[label]
                del self.has_labels[phrase_string]

    def is_label(self, label: str) -> bool:
        """Check if label is registered as label of any known phrase.

        :param label: a label string to be checked
        :type label: str
        :return: a boolean whether the label belongs to a known phrase
        :rtype: bool
        """
        return label in self.is_label_of

    def has_label(self, phrase_string: str) -> bool:
        """Check if a registered phrase has a label.

        :param phrase_string: a phrase string of a registered phrase
        :type phrase_string: str
        :return: a boolean indicating if the registered phrase has a label
        """
        return phrase_string in self.has_labels

    def get_labels(self, phrase: Union[str, Phrase]) -> Set[str]:
        """Return the label(s) of a registered phrase.

        :param phrase: a phrase string or object
        :type phrase: Union[str, Phrase]
        :return: a set of labels
        :rtype: List[str]
        """
        phrase_string = phrase if isinstance(phrase, str) else phrase.phrase_string
        if not self.has_phrase(phrase_string):
            raise KeyError(f"Unknown phrase: {phrase}")
        return self.has_labels[phrase]

    def validate_entry_phrase(self, entry: Dict[str, Union[str, int, float, list]]) -> None:
        """Check if a given phrase (as dictionary) is registered.

        :param entry: a phrase dictionary with a 'phrase' property
        :type entry: Dict[str, Union[str, int, float, list]]
        """
        if 'phrase' not in entry:
            raise KeyError("Keywords as list of dictionaries should have 'phrase' property")
        if not isinstance(entry['phrase'], str):
            raise ValueError("phrase property must be a string")
        if entry['phrase'] not in self.phrase_index:
            raise KeyError("Unknown phrase")

    def add_custom(self, custom: List[Dict[str, Union[str, int, float, list]]]) -> None:
        """Add custom key/value pairs to the entry as phrase metadata.

        param entry: an Array of phrase dictionaries, each with a 'phrase' property and additional key/value pairs
        type entry: Dict[str, Union[str, int, float, list]]
        """
        for entry in custom:
            if not isinstance(entry, dict):
                continue
            self.validate_entry_phrase(entry)
            # make sure the custom entry is a copy of the original and not a reference to the same object
            self.custom[entry['phrase']] = copy.deepcopy(entry)

    def remove_custom(self, custom: List[Dict[str, any]]) -> None:
        """Remove custom properties for a list of phrases.

        :param custom: a list of phrase dictionaries with custom properties to remove
        :type custom: List[Dict[str, any]]
        """
        for entry in custom:
            self.validate_entry_phrase(entry)
            for custom_property in entry:
                del self.custom[entry['phrase']][custom_property]

    def has_custom(self, phrase_string: str, custom_property: str) -> bool:
        """Check if a phrase has a given custom property.

        :param phrase_string: a phrase string of a registered phrase.
        :type phrase_string: str
        :param custom_property: the name of a custom property of the registered phrase
        :type custom_property: str
        :return: a boolean to indicate whether the phrase has a custom property of the given property name
        :rtype: bool
        """
        return phrase_string in self.custom and custom_property in self.custom[phrase_string]

    def get(self, phrase_string: str, custom_property: str) -> any:
        """Get the value of a custom property for a given phrase.

        :param phrase_string: a phrase string of a registered phrase.
        :type phrase_string: str
        :param custom_property: the name of a custom property of the registered phrase
        :type custom_property: str
        :return: the custom property of a given phrase
        :rtype: any
        """
        if phrase_string not in self.phrase_index:
            raise KeyError("Unknown phrase_string")
        if not self.has_custom(phrase_string, custom_property):
            raise ValueError("Unknown custom property")
        return self.custom[phrase_string][custom_property]

    def index_phrase_words(self, phrase: Phrase) -> None:
        """Index a phrase on its individual words, for exact match look up routines.

        :param phrase: a phrase object that is part of the phrase model
        :type phrase: Phrase
        """
        if phrase.phrase_string not in self.phrase_type:
            raise ValueError(f"Cannot index phrase words for non-registered phrase: {phrase.phrase_string}")
        for wi, word in enumerate(re.finditer(r"\w+", phrase.phrase_string)):
            if wi == 0:
                self.first_word_in_phrase[word.group(0)][phrase.phrase_string] = word.start()
            self.word_in_phrase[word.group(0)].add(phrase.phrase_string)

    def remove_phrase_words(self, phrase: Phrase) -> None:
        """Remove the individual words of a phrase from the index. Only use this is you are removing
        the phrase from the phrase model.

        :param phrase: a phrase object that is part of the phrase model
        :type phrase: Phrase
        """
        for wi, word in enumerate(re.finditer(r"\w+", phrase.phrase_string)):
            if wi == 0:
                del self.first_word_in_phrase[word.group(0)][phrase.phrase_string]
                if len(self.first_word_in_phrase[word.group(0)].keys()) == 0:
                    del self.first_word_in_phrase[word.group(0)]
            self.word_in_phrase[word.group(0)].remove(phrase.phrase_string)
            if len(self.word_in_phrase[word.group(0)]) == 0:
                del self.word_in_phrase[word.group(0)]
